/**
 * Material swapper component: use this component to swap out materials on the parent's mesh.
 * Disposes memory on destroy if necessary. Restores the materials if destroyed.
 * @moduleName MaterialSwapper
 * @snippet MaterialSwapper|Constructor
MaterialSwapper({
    id: '$1', // unique id for refcounting
    object3D: null, // the target mesh, otherwise the parent's Object3D
    deepSwap: true, // swaps out all children
    type: 'MeshBasicMaterial',
    materialSettings: {}
})
 */
bento.define('components/materialswapper', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'onigiri/refcount'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    RefCount
) {
    'use strict';
    return function (settings) {
        var entity;
        var object3D = settings.object3D || settings.mesh;
        var id = settings.id || '';
        var type = settings.type || 'MeshBasicMaterial';
        var deepSwap = settings.deepSwap || false;
        var materialSettings = settings.materialSettings || {};
        var refCount = new RefCount({
            id: id,
            constructor: function () {
                // the reference counted asset to make
                return new THREE[type](materialSettings);
            }
        });
        var oldMaterials = [];
        var component = new Object({
            name: 'materialSwapper',
            material: null,
            start: function (data) {
                var material = refCount.useAsset().asset;
                if (!object3D) {
                    console.warn('Warning: no Object3D found in MaterialSwapper');
                    return;
                }
                if (deepSwap) {
                    object3D.traverse(function (child) {
                        if (child.material) {
                            oldMaterials.push({
                                object: child,
                                material: child.material
                            });
                            child.material = material;
                        }
                    });
                } else {
                    if (object3D && object3D.material) {
                        oldMaterials.push({
                            object: object3D,
                            material: object3D.material
                        });
                        object3D.material = material;
                    }
                }
            },
            destroy: function (data) {
                // restore old materials
                Utils.forEach(oldMaterials, function (item) {
                    item.object.material = item.material;
                });
                refCount.dispose();
            },
            attached: function (data) {
                entity = data.entity;
                if (!object3D) {
                    // use the parent entity's object3D
                    object3D = entity.object3D;
                }
            },
            setObject3D: function (target) {
                object3D = target;
                return component;
            }
        });
        return component;
    };
});