/**
 * Makes an entity with a body follow a path
 * @moduleName PathFollower
 */
bento.define('components/pathfollower', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals',
    'onigiri/onigiri'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals,
    Onigiri
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        // var viewport = Bento.getViewport();
        var pathName = settings.pathName;
        var entity;

        /* properties that we get from the path */

        var levelScale;
        var points; // array of points e.g. [ {x:10,y:20} , {x:5,y:3} ]
        var pixelsPerSecond;
        var timePerPoint;
        var ease;

        /* variables for movement logic */
        var startPos;
        var currentPointIndex = 0;
        var currentPoint;
        var nextPointIndex;
        var nextPoint;
        var closedLoop = false; // we move back and forth unless the final point is the same as the first point
        var reversing = false; // are we moving backwards along the path?

        var follow;
        var body;

        var component = {
            name: 'pathFollower',
            start: function (data) {
                startPos = entity.position.clone();
                body = entity.getComponent('body');
            },
            destroy: function (data) {},
            update: function (data) {
                if (follow) {
                    follow.update({
                        speed: data.speed,
                    });
                }
                if (!points) {
                    var path = Globals.paths[pathName];
                    if (!path) {
                        return;
                    }

                    points = path.points;
                    pixelsPerSecond = path.pixelsPerSecond; // no standard value for this, setting this will overule the tween so it follows the path at a steady place
                    timePerPoint = path.timePerPoint || 120;
                    ease = path.ease || 'linear';
                    levelScale = path.levelScale || 1;

                    currentPointIndex = 0;
                    currentPoint = points[0];
                    nextPoint = points[1];

                    if (points[0].x === points[points.length - 1].x && points[0].y === points[points.length - 1].y) {
                        closedLoop = true;
                    }
                    moveToNextPoint();
                }
            },
            attached: function (data) {
                entity = data.entity;
            }
        };

        var tweenUpdate = function (v, t) {
            var pathX = currentPoint.x + (nextPoint.x - currentPoint.x) * v;
            var pathY = currentPoint.y + (nextPoint.y - currentPoint.y) * v;

            // sometimes tiled polylines are not normalised... so we should subtract the first point
            pathX -= points[0].x;
            pathY -= points[0].y;

            var x1 = entity.position.x;
            var y1 = entity.position.y;
            var x2 = startPos.x + pathX * levelScale * Onigiri.pxToThree;
            var y2 = startPos.y + pathY * levelScale * Onigiri.pxToThree * -1;

            if (!body) {
                body = entity.getComponent('body');
                if (!body) {
                    return;
                }
            }

            var type = body.getType();
            var f = new Vector2(x2 - x1, y2 - y1);
            if (type !== 'static') {
                f.scale(60); // multiply to a second to apply proper forces
                body.setLinearVelocity(f);
            } else {
                body.setPosition(body.getPosition().addTo(f));
            }
        };


        var moveUpdate = function (data) {
            var moveAmount = pixelsPerSecond * data.speed / 60;
            var pathVector = new Vector2(0, 0);
            var magnitude = 0;
            var v = 0;
            while (moveAmount > 0) {
                pathVector.x = nextPoint.x - currentPoint.x;
                pathVector.y = nextPoint.y - currentPoint.y;
                magnitude = pathVector.magnitude();
                this.pathProgress += moveAmount;
                moveAmount = Math.max(this.pathProgress - magnitude, 0);
                if (moveAmount) {
                    currentPointIndex = nextPointIndex;
                    currentPoint = nextPoint;
                    moveToNextPoint();
                } else {
                    v = this.pathProgress / magnitude;
                }
            }

            var pathX = currentPoint.x + (nextPoint.x - currentPoint.x) * v;
            var pathY = currentPoint.y + (nextPoint.y - currentPoint.y) * v;

            // add levelScale
            pathX *= levelScale;
            pathY *= levelScale;

            // Switch from pixel coordinates to three coordinate
            pathX *= Onigiri.pxToThree;
            pathY *= Onigiri.pxToThree;

            // flip Y because of Three coordinate system
            pathY = -pathY;

            var x1 = entity.position.x;
            var y1 = entity.position.y;
            var x2 = startPos.x + pathX;
            var y2 = startPos.y + pathY;

            if (!body) {
                body = entity.getComponent('body');
                if (!body) {
                    return;
                }
            }

            var type = body.getType();
            var f = new Vector2(x2 - x1, y2 - y1);
            if (type !== 'static') {
                f.scale(60); // multiply to a second to apply proper forces
                body.setLinearVelocity(f);
            } else {
                body.setPosition(body.getPosition().addTo(f));
            }
        };

        // figure out the next point
        // start the tween to go there
        // when we're done, the current point becomes the next point
        // then we repeat the process.
        var moveToNextPoint = function () {
            if (!reversing) {
                nextPointIndex = currentPointIndex + 1;
                if (closedLoop) {
                    if (nextPointIndex >= points.length - 1) {
                        nextPointIndex = 0;
                    }
                } else {
                    if (nextPointIndex >= points.length) {
                        reversing = true;
                        nextPointIndex = currentPointIndex - 1;
                    }
                }
            } else {
                nextPointIndex = currentPointIndex - 1;
                if (nextPointIndex < 0) {
                    nextPointIndex = 1;
                    reversing = false;
                }
            }
            nextPoint = points[nextPointIndex];

            if (isNaN(pixelsPerSecond)) {
                follow = new Tween.TweenBehavior({
                    from: 0,
                    to: 1,
                    in: timePerPoint,
                    ease: ease,
                    onUpdate: tweenUpdate,
                    onComplete: function () {
                        currentPointIndex = nextPointIndex;
                        currentPoint = nextPoint;
                        moveToNextPoint();
                    }
                });
            } else {
                follow = follow || new Object({
                    pathProgress: 0,
                    update: moveUpdate
                });
                follow.pathProgress = 0;
            }
        };

        return component;
    };
});