/**
 * Swaps textures in a model
 * @moduleName TextureSwapper
 * snippet TextureSwapper|constructor
TextureSwapper({
    mesh: ${1:mesh}, // THREE.Mesh
    perInstance: true, // clones material per instance if true
    deepSwap: true,
    onTexture: function (oldTexture) {
        // example:
        var newTexture = Bento.assets.getTexture('$2');
        return newTexture;
    }
})
 * @snippet TextureSwapper()|TextureSwapper
TextureSwapper({
    mesh: ${1:mesh}, 
    imageName: '${2:imageName}'
})
 */
bento.define('components/textureswapper', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'globals',
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Globals
) {
    'use strict';
    // shared materials cache
    var materials = {};
    var TextureSwapper = function (settings) {
        var perInstance = Utils.getDefault(settings.perInstance, true);
        var mesh = settings.mesh;
        var onTexture = settings.onTexture;
        var deepSwap = Utils.getDefault(settings.deepSwap, true);
        var allMaterials = settings.allMaterials;
        var entity;
        var onChild = function (child) {
            var material;
            var swapped;
            var newTexture;
            var oldMaterial;
            var id;
            // we can't just swap out the texture in the material, 
            // because all meshes that use this material will also swap their texture
            // instead we swap out the material
            if (child.material && (child.material.map || allMaterials)) {
                oldMaterial = child.material.map;
                newTexture = onTexture(oldMaterial);
                if (!newTexture) {
                    // nothing to swap
                    return;
                }

                // copy texture settings (particularly encoding)
                if (oldMaterial) {
                    newTexture.magFilter = oldMaterial.magFilter;
                    newTexture.minFilter = oldMaterial.minFilter;
                    newTexture.wrapS = oldMaterial.wrapS;
                    newTexture.wrapT = oldMaterial.wrapT;
                    newTexture.encoding = oldMaterial.encoding;
                }

                component.texture = newTexture;

                if (!perInstance) {
                    // swap (shared) material's texture
                    child.material.map = newTexture;
                    child.material.emissiveMap = newTexture;
                    child.material.needsUpdate = true;
                    return;
                }

                material = child.material;
                // check if it exists in cache
                id = material.uuid + '_' + newTexture.uuid;
                swapped = materials[id];

                if (!swapped) {
                    // initialize material if non-existent
                    // child.material = child.material.clone();
                    child.material = new THREE.MeshBasicMaterial();
                    child.material.map = newTexture;
                    // child.material.emissiveMap = newTexture;
                    child.material.needsUpdate = true;

                    // store result as original uuid + '_'
                    materials[id] = child.material;
                    // TODO: refcount material
                    swapped = child.material;
                }

                child.material = swapped;

                if (allMaterials) {
                    child.material.skinning = true;
                }

            }
        };
        var component = new Object({
            name: 'textureSwapper',
            texture: null,
            start: function (data) {
                if (!mesh) {
                    // try to find it?
                    entity.getComponent('mesh', function (meshComponent) {
                        mesh = meshComponent.mesh;
                    });
                    if (!mesh) {
                        console.log('ERROR: could not find mesh');
                        return;
                    }
                }
                if (deepSwap) {
                    mesh.traverse(onChild);
                } else {
                    onChild(mesh);
                }
            },
            attached: function (data) {
                entity = data.entity;
            }
        });
        return component;
    };
    TextureSwapper.clearCache = function () {
        Utils.forEach(materials, function (material) {
            if (material.dispose) {
                material.dispose();
            }
        });
        materials = {};
    };

    TextureSwapper.UA = function (settings) {
        var mesh = settings.mesh;
        var deepSwap = settings.deepSwap;
        var imageName = settings.imageName;
        var allMaterials = settings.allMaterials;

        return new TextureSwapper({
            mesh: mesh, // THREE.Mesh
            perInstance: false, // clones material per instance if true
            deepSwap: deepSwap,
            allMaterials: allMaterials,
            onTexture: function (oldTexture) {
                var texture;
                var imagePack;

                if (Bento.assets.hasTexture(imageName)) {
                    // check if it was loaded before
                    texture = Bento.assets.getTexture(imageName);
                } else {
                    // load from image
                    imagePack = Bento.assets.getImage(imageName);
                    if (!imagePack) {
                        return;
                    }
                    texture = imagePack.image.texture || new THREE.Texture(imagePack.image);
                    imagePack.image.texture = texture;
                }

                if (texture === oldTexture) {
                    return;
                }
                texture.flipY = false;
                if (oldTexture) {
                    texture.flipY = oldTexture.flipY;
                }

                texture.needsUpdate = true;
                return texture;
            }
        });
    };

    return TextureSwapper.UA;
});