/**
 * Basket ball
 * @moduleName BasketBall
 * @snippet BasketBall|Constructor
BasketBall({})
 */
bento.define('entities/basketball', [
    'bento',
    'bento/utils',
    'bento/entity',
    'bento/math/vector2',
    'bento/math/rectangle',
    'onigiri/entity3d',
    'onigiri/primitive',
    'onigiri/onigiri',
    'modules/box2d',
    'onigiri/clickcaster',
    'bento/components/clickable',
    'bento/tween',
    'components/targetsize',
    'components/materialswapper',
    'components/trail',
    'managers/savevalues',
    'globals',
    'bento/components/eventlistener'
], function (
    Bento,
    Utils,
    Entity,
    Vector2,
    Rectangle,
    Entity3D,
    Primitive,
    Onigiri,
    Box2D,
    ClickCaster,
    Clickable,
    Tween,
    TargetSize,
    MaterialSwapper,
    Trail,
    SaveValues,
    Globals,
    EventListener
) {
    'use strict';
    var radius = 1;
    var BallGraphic = function (settings) {
        var meshName = settings.meshName;
        var ballMesh = Onigiri.getMesh(meshName || 'balls/ball-' + SaveValues.currentBall + '.gltf');
        var ballSettings = Utils.getSafe(Globals, ['ballSettings', SaveValues.currentBall], {});
        var sizeSettings = (function () {
            var fitBy = ballSettings.fitBy || 'width';
            var setting = {};
            setting[fitBy] = radius * 2;
            return setting;
        })();
        var ball = new Entity3D({
            name: 'ball',
            object3D: ballMesh, // THREE.object3D reference
            castShadow: true,
            components: [
                new TargetSize(sizeSettings),
                new MaterialSwapper({
                    id: 'ball-' + SaveValues.currentBall, // unique id for refcounting
                    type: 'MeshPhongMaterial',
                    deepSwap: true,
                    materialSettings: {
                        map: Onigiri.getTexture('textures/balls/ball-' + SaveValues.currentBall, {
                            wrapS: Utils.getDefault(ballSettings.wrapS, THREE.RepeatWrapping),
                            wrapT: Utils.getDefault(ballSettings.wrapT, THREE.RepeatWrapping),
                            flipY: Utils.getDefault(ballSettings.flipY, false)
                        }),
                        specular: ballSettings.specular || 0xdd7f56,
                        shininess: Utils.getDefault(ballSettings.shininess, 8),
                        reflectivity: Utils.getDefault(ballSettings.reflectivity, 0.5),
                    }
                })
            ].concat(settings.components || [])
        });
        return ball;
    };
    var BasketBall = function (settings) {
        var bodyDynamic = new Box2D.Body({
            type: 'dynamic',
            fixture: {
                shape: new Box2D.CircleShape(radius, new Vector2(0, 0)),
                friction: 0.2,
                restitution: 0.8,
                density: 50,
                allowSleep: false,
                onBeginContact: function (other, contact) {
                    if (!other.isSensor || entity.name === 'basketHoop') {
                        entity.onContact(other, contact);
                    }
                },
            }
        });
        var propellor = new Entity3D({
            name: 'propellor',
            object3D: Onigiri.getMesh('propeller.gltf'), // THREE.object3D reference
            position: new THREE.Vector3(0, radius / 2, 0),
            components: [
                new MaterialSwapper({
                    id: 'propellor', // unique id for refcounting
                    type: 'MeshPhongMaterial',
                    deepSwap: true,
                    materialSettings: {
                        color: 0xffffff,
                        specular: 0xffffff,
                        shininess: 8,
                        reflectivity: 0.5,
                    }
                }),
                new Object({
                    name: 'behavior',
                    update: function (data) {
                        propellor.euler.y = propellor.timer / 2;

                        if (entity.isFollowingPath) {
                            propellor.visible = true;
                            entity.euler.x = 0;
                            entity.euler.y = 0;
                            entity.euler.z = 0;
                            ball.euler.z = Math.PI / 10 * Math.sin(ball.timer / 20);
                        } else {
                            propellor.visible = false;
                        }
                    }
                })
            ]
        });
        var trailTarget = new THREE.Object3D();
        var trail = new Trail({
            mesh: trailTarget,
            size: 0.5,
            trailLength: 16
        });
        var ball = new BallGraphic({
            components: [
                propellor
            ]
        });
        var skinListener = new EventListener({
            name: 'skinListener',
            eventName: 'refreshBallSkin',
            ignorePause: true,
            onEvent: function (data) {
                ball.remove(propellor);
                ball.removeSelf();
                ball = new BallGraphic({
                    components: [
                        propellor
                    ]
                });
                entity.attach(ball);
            }
        });
        var entity = new Entity3D({
            name: 'ball',
            object3D: trailTarget,
            position: settings.position,
            family: ['balls'],
            components: [
                ball,
                skinListener,
                trail
            ]
        }).extend({
            body: bodyDynamic,
            player: null,
            isFollowingPath: false,
            // overwrite this
            onContact: function () {},
            setDynamic: function () {
                if (!bodyDynamic.parent) {
                    entity.attach(bodyDynamic);
                }
            },
            setKinematic: function (onComplete) {
                if (bodyDynamic.parent) {
                    Box2D.defer(function () {
                        entity.remove(bodyDynamic);
                        if (onComplete) {
                            onComplete();
                        }
                    });
                }
            },
            shake: function () {
                new Tween({
                    from: 0.2,
                    to: 0,
                    in: 10,
                    onUpdate: function (v, t) {
                        ball.position.x = Utils.getRandomRangeFloat(-v, v);
                        ball.position.y = Utils.getRandomRangeFloat(-v, v);
                    },
                    onComplete: function () {
                        ball.position.x = 0;
                        ball.position.y = 0;
                    }
                });
            },
            snapTrail: function () {
                trail.snap();
            }
        });
        return entity;
    };
    BasketBall.BallGraphic = BallGraphic;
    return BasketBall;
});