/**
 * Draw controller, combines basket ball & drawplane controls
 * @moduleName DrawController
 * @snippet DrawController|Constructor
DrawController({})
 */
bento.define('entities/drawcontroller', [
    'bento',
    'bento/utils',
    'bento/entity',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/tween',
    'onigiri/onigiri',
    'modules/box2d',
    'onigiri/spriteplane',
    'onigiri/entity3d',
    'bento/eventsystem',
    'globals',
    'bento/components/spritecontainer',
    'entities/drawplane',
    'bento/components/eventlistener',
    'bento/components/clickable'
], function (
    Bento,
    Utils,
    Entity,
    Vector2,
    Rectangle,
    Tween,
    Onigiri,
    Box2D,
    SpritePlane,
    Entity3D,
    EventSystem,
    Globals,
    SpriteContainer,
    DrawPlane,
    EventListener,
    Clickable
) {
    'use strict';
    var DrawController = function (settings) {
        var ball = settings.ball;
        var hoop = settings.hoop;
        var drawPlane = settings.drawPlane || new DrawPlane(settings);
        var travelDistance = 0;
        var totalDistance = 0;
        var travelSpeed = 2.5 / 6 / 3;
        var targetTravelSpeed = 2.5 / 6;
        var isMoving = true;
        var finalAngle = 0;
        var targetPosition;
        var travelArray = [];
        var autoAimThreshold = 75; // percent
        var uaHand;
        var behavior = new Object({
            name: 'behavior',
            start: function (data) {
                if (!ball) {
                    ball = Bento.objects.get('ball');
                }
                if (!hoop) {
                    hoop = Bento.objects.get('basketHoop');
                }
                if (!drawPlane) {
                    drawPlane = Bento.objects.get('drawPlane');
                } else if (!drawPlane.isAdded) {
                    Bento.objects.attach(drawPlane);
                }
                if (drawPlane && ball && hoop) {
                    ball.onContact = function (other, contact) {
                        // force end
                        var manifold;
                        if (targetPosition) {
                            behavior.endTravel();
                            manifold = contact.GetManifold();

                            if (entity.name !== 'basketHoop') {
                                ball.shake();
                                Bento.objects.attach(new Entity3D({
                                    name: 'hitEffect',
                                    position: ball.position.clone().add(new THREE.Vector3(
                                        manifold.m_localPoint.x * Onigiri.pxToThree,
                                        manifold.m_localPoint.y * -Onigiri.pxToThree,
                                        0
                                    )),
                                    components: [
                                        new SpritePlane({
                                            width: 3,
                                            imageName: 'fx/hit',
                                            padding: 1,
                                            frameCountX: 2,
                                            frameCountY: 2,
                                            animations: {
                                                hit: {
                                                    frames: [0, 1, 2, 3],
                                                    speed: 0.4
                                                }
                                            }
                                        }),
                                        new Object({
                                            name: 'behavior',
                                            start: function () {
                                                var entity = this.parent;
                                                entity.getComponent('sprite3D', function (sprite3D) {
                                                    sprite3D.sprite.setAnimation('hit', function () {
                                                        entity.removeSelf();
                                                    });
                                                });
                                            }
                                        })
                                    ]
                                }));
                            }
                        }
                    };

                    // overwrite
                    drawPlane.onDrawStart = function () {
                        pulseBehavior.shouldPulse = false;
                        drawPlane.drawFrom(0);

                        EventSystem.fire('pathDrawingStarted');

                        isMoving = true;

                        // remove any tutorial
                        Bento.objects.get('drawTutorial', function (drawTutorial) {
                            drawTutorial.removeSelf();
                        });

                        if (!uaHand && Bento.saveState.load('cheat-showUaHand', false)) {
                            createUaHand();
                        }
                    };
                    drawPlane.onDraw = function (array, distance, point) {
                        if (uaHand) {
                            var pos = Onigiri.projectToBento(point.clone());
                            uaHand.position.x = pos.x - 12;
                            uaHand.position.y = pos.y - 4;
                            uaHand.visible = true;
                        }
                    };
                    drawPlane.onDrawEnd = function (lineArray, lastAngle, distanceDrawn) {

                        // start traveling through the path
                        if (lineArray.length > 1) {
                            travelDistance = 0;
                            totalDistance = distanceDrawn;
                            finalAngle = lastAngle;
                            travelArray = lineArray;

                            if (ball.player && ball.player.shoot) {
                                ball.player.shoot();
                            }
                            ball.setDynamic();

                            // snap to first point?
                            ball.body.setPosition(new Vector2(lineArray[0].x, lineArray[0].y));

                            // allow drawing only once (?)
                            drawPlane.active = false;
                        }

                        if (uaHand) {
                            uaHand.visible = false;
                        }
                    };
                }
            },
            update: function (data) {
                if (travelArray.length) {
                    // take a step
                    if (isMoving) {
                        travelDistance += travelSpeed * data.speed;
                        travelSpeed = Utils.approach(travelSpeed, targetTravelSpeed, targetTravelSpeed / 60);
                    }

                    // find interpolated position
                    var traveled = 0;
                    var lastPoint = travelArray[0];
                    var lastIndex = 0;
                    Utils.forEach(travelArray, function (point, i, l, breakLoop) {
                        var diff = point.clone().sub(lastPoint);
                        var distanceBetweenPoints = lastPoint.distanceTo(point);
                        var interpolation;

                        lastIndex = i - 1;

                        // will reach target between these points, find the interpolated position
                        if (traveled + distanceBetweenPoints >= travelDistance) {
                            interpolation = (travelDistance - traveled) / distanceBetweenPoints;
                            lastPoint = lastPoint.clone().add(diff.multiplyScalar(interpolation));
                            breakLoop();
                            return;
                        }

                        traveled += distanceBetweenPoints;
                        lastPoint = point;

                        if (i === l - 1) {
                            // reached the end of array: nothing left to interpolate
                            travelDistance = totalDistance;
                        }
                    });

                    drawPlane.drawFrom(travelDistance);

                    // ball.body.setPosition(new Vector2(lastPoint.x, lastPoint.y));
                    targetPosition = lastPoint;
                    ball.isFollowingPath = true;

                    if (travelDistance >= totalDistance) {
                        // snap to final point
                        ball.body.setPosition(new Vector2(lastPoint.x, lastPoint.y));
                        behavior.endTravel();
                    }
                }

                if (targetPosition) {
                    // interpolate towards position
                    var currentPosition = ball.body.getPosition();

                    // set speed to the position difference
                    var positionDiff = new Vector2(targetPosition.x, targetPosition.y).subtract(currentPosition);

                    // convert back to bento coords
                    positionDiff.scalarMultiplyWith(Onigiri.threeToPx);

                    // move by setting velocity
                    ball.body.setAwake(true);
                    ball.body.setLinearVelocity(positionDiff);
                }
            },
            endTravel: function (noRelease) {
                travelArray = [];
                targetPosition = null;

                if (!noRelease) {
                    EventSystem.fire('releasedBall');
                }

                // hide plane
                drawPlane.visible = false;
                ball.isFollowingPath = false;
                drawPlane.drawFrom(Infinity);

                if (Bento.saveState.load('cheat-noAutoAim', false)) {
                    return;
                }
                // TODO: explicitly set velocity to follow a prediction path
                // var velocity = new Vector2(travelSpeed * Onigiri.threeToPx, 0);
                // velocity.rotateRadian(finalAngle);
                // ball.body.setLinearVelocity(velocity);
                // console.log(velocity.x, velocity.y);

                autoAim();
            }
        });
        var pulseBehavior = new Object({
            name: 'behavior',
            shouldPulse: true,
            pulseTimer: 0,
            update: function (data) {
                pulseBehavior.pulseTimer += data.speed;
                if (pulseBehavior.shouldPulse && pulseBehavior.pulseTimer > 60 && ball) {
                    pulseBehavior.pulseTimer = 0;
                    pulseBehavior.pulseBall();
                }
            },
            pulseBall: function () {
                var haloSprite = new SpritePlane({
                    width: 2,
                    imageName: 'fx/ball-halo',
                    originRelative: new Vector2(0.5, 0.5)
                });
                var effect = new Entity3D({
                    name: 'pulseEffect',
                    position: ball.position.clone(),
                    components: [
                        haloSprite
                    ]
                });
                Bento.objects.attach(effect);

                new Tween({
                    from: 1,
                    to: 2,
                    in: 30,
                    ease: 'exponential',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        effect.scale.x = v;
                        effect.scale.y = v;
                        effect.scale.z = v;
                    },
                    onComplete: function () {
                        effect.removeSelf();
                    }
                });
                new Tween({
                    from: 1,
                    to: 0,
                    in: 30,
                    onUpdate: function (v, t) {
                        haloSprite.sprite.material.opacity = v;
                    }
                });
            }
        });
        var autoAim = function () {
            // TODO: go through all possible targets, target the closest one ...?

            var velocity = ball.body.getLinearVelocity();

            var det;
            var detSqrt;
            var dy = (hoop.position.y - ball.position.y);
            var dx = (hoop.position.x - ball.position.x);
            // convert velocity & gravity to bento 
            var velX = velocity.x / 60;
            var velY = velocity.y / -60;
            var gravity = -Box2D.gravity.y / -60 / 60;
            var correctionPercent = 0;

            // 
            det = velY * velY + 2 * gravity * dy;
            detSqrt = Math.sqrt(Math.abs(det));
            velX = gravity * dx / (detSqrt - velY);

            velX = -velX * 60;

            correctionPercent = ((velX - velocity.x) / velocity.x * 100);

            console.log('Ball correction: ' + ((velX - velocity.x) / velocity.x * 100).toFixed(1) + '%');

            if (correctionPercent < -autoAimThreshold || correctionPercent > autoAimThreshold || isNaN(correctionPercent) || !isFinite(correctionPercent)) {
                // correction is too big: cancel out
                return;
            }

            // correct ball
            velocity.x = velX;
            ball.body.setLinearVelocity(velocity);
        };
        var createUaHand = function () {
            uaHand = new Entity({
                z: Globals.Layers.GUI + 1,
                name: 'uaHand',
                rotation: -Math.PI / 8,
                visible: false,
                components: [
                    new SpriteContainer({
                        width: 96,
                        imageName: 'gui/ua-hand',
                        originRelative: new Vector2(0, 0),
                    }),
                    new Object({
                        name: 'behavior',
                        update: function (data) {},
                    })
                ]
            });
            Bento.objects.attach(uaHand);
        };
        var caughtListener = new EventListener({
            name: 'caughtListener',
            eventName: 'caughtBall',
            ignorePause: false,
            onEvent: function (data) {
                // instantly stop ball travel
                behavior.endTravel(true);

                // reset
                travelDistance = 0;
                totalDistance = 0;
                travelArray = [];
                targetPosition = null;

                // re-activate drawplane
                drawPlane.active = true;
                drawPlane.visible = true;
                drawPlane.reset();

                pulseBehavior.shouldPulse = true;
                pulseBehavior.pulseTimer = 30;

            }
        });

        var style = Globals.tapStyle;
        var pointerDown = null;
        var clickable = new Clickable({
            pointerDown: function (data) {
                if (style === 'tap') {
                    if (travelArray.length) {
                        // tap style = swapping state
                        isMoving = !isMoving;
                        if (!isMoving) {
                            pulseBehavior.shouldPulse = true;
                            pulseBehavior.pulseTimer = 60;
                            addPauseIcon();
                            EventSystem.fire('pauseBall');
                        } else {
                            pulseBehavior.shouldPulse = false;
                            pulseBehavior.pulseTimer = 30;
                            EventSystem.fire('resumeBall');
                        }
                    }
                } else if (style === 'tapAndHold') {
                    if (pointerDown === null) {
                        pointerDown = data.id;
                        isMoving = false;
                        if (travelArray.length) {
                            pulseBehavior.shouldPulse = true;
                            pulseBehavior.pulseTimer = 60;
                            addPauseIcon();
                            EventSystem.fire('pauseBall');
                        }
                    }
                }
            },
            pointerUp: function (data) {
                if (style === 'tapAndHold' && pointerDown === data.id) {
                    pointerDown = null;
                    isMoving = true;
                    if (travelArray.length) {
                        pulseBehavior.shouldPulse = false;
                        pulseBehavior.pulseTimer = 30;
                        EventSystem.fire('resumeBall');
                    }
                }
            }
        });

        var entity = new Entity({
            name: 'drawController',
            components: [
                behavior,
                pulseBehavior,
                caughtListener,
                clickable
            ]
        });
        var addPauseIcon = function () {
            var sprite = new SpritePlane({
                width: 1.25,
                imageName: 'gui/icons/pause',
                originRelative: new Vector2(0.5, 0.5),
            });
            var pauseIcon = new Entity3D({
                name: 'pauseIcon',
                components: [
                    sprite,
                    new Object({
                        name: 'behavior',
                        start: function (data) {},
                        destroy: function (data) {},
                        update: function (data) {
                            // follow ball
                            pauseIcon.position.x = ball.position.x;
                            pauseIcon.position.y = ball.position.y;
                            pauseIcon.position.z = 1;

                            sprite.visible = (Math.floor(pauseIcon.ticker / 20) % 2) === 0;
                        }
                    }),
                    new EventListener({
                        name: 'resumeListener',
                        eventName: 'resumeBall',
                        ignorePause: false,
                        onEvent: function (data) {
                            pauseIcon.removeSelf();
                        }
                    })
                ]
            });
            Bento.objects.attach(pauseIcon);
        };

        return entity;
    };
    return DrawController;
});