/**
 * Invisible plane to raycast on
 * @moduleName DrawPlane
 * @snippet DrawPlane|Constructor
DrawPlane({})
 */
bento.define('entities/drawplane', [
    'bento',
    'bento/utils',
    'onigiri/entity3d',
    'onigiri/primitive',
    'onigiri/clickcaster',
    'onigiri/onigiri',
    'components/line',
    'bento/eventsystem',
    'components/healthbar3d',
    'bento/components/eventlistener'
], function (
    Bento,
    Utils,
    Entity3D,
    Primitive,
    ClickCaster,
    Onigiri,
    Line,
    EventSystem,
    HealthBar3D,
    EventListener
) {
    'use strict';
    var invisibleMaterial = new THREE.MeshBasicMaterial({
        transparent: true,
        opacity: 0
    });
    var ballMaterial = new THREE.MeshBasicMaterial({
        color: '#b95b01',
    });
    var Particle = function (settings) {
        var position = settings.position;
        var velocity = settings.velocity || new THREE.Vector3(0, 0, 0);
        var gravity = -0.01;
        var scale;
        var lifeTime = 30;
        var behavior = new Object({
            name: 'behavior',
            start: function (data) {
                position = entity.position;
                scale = entity.scale;
            },
            destroy: function (data) {},
            update: function (data) {
                velocity.y += gravity;
                position.x += velocity.x;
                position.y += velocity.y;
                position.z += velocity.z;

                scale.x = Utils.approach(scale.x, 0, 1 / lifeTime);
                scale.y = scale.x;
                scale.z = scale.x;

                if (entity.timer > lifeTime) {
                    entity.removeSelf();
                }
            }
        });
        var entity = new Primitive({
            name: 'particle',
            shape: 'sphere',
            parameters: [
                0.1, // radius
                4, // widthSegments
                4 // heightSegments
            ],
            material: ballMaterial,
            disposeGeometry: true,
            disposeMaterial: false,
            position: position,
            components: [
                behavior
            ]
        });
        return entity;
    };
    var DrawPlane = function (settings) {
        var drawDistanceStart = settings.drawDistanceStart || 3;
        var plane = new Primitive({
            shape: 'plane',
            parameters: [
                1000, // width
                1000 // height
            ],
            material: invisibleMaterial
        });
        var pointerId = null;
        var previousPoint;
        var lastAngle;
        var lineComponent = new Line({});
        var healthbarTarget = {
            health: 100,
            position: new THREE.Vector3(0, 0, 0)
        };
        var healthbar = new HealthBar3D({
            height: 2.5,
            width: 2.5 * 1.5,
            targetEntity: healthbarTarget
        });
        var lineArray = [];
        var distanceDrawn = 0;
        var players = Bento.objects.getByFamily('players');
        var behavior = new Object({
            name: 'behavior',
            update: function (data) {
                if (entity.cancelFrame) {
                    entity.cancelFrame -= 1;
                }
            }
        });
        var cancelListener = new EventListener({
            name: 'cancelListener',
            eventName: 'cancelThrow',
            ignorePause: true,
            onEvent: function (data) {
                entity.cancelFrame = 2;
            }
        });
        var entity = new Entity3D({
            name: 'drawPlane',
            position: new THREE.Vector3(0, 0, 0),
            components: [
                plane,
                lineComponent,
                healthbar,
                cancelListener,
                new ClickCaster({
                    raycastMesh: plane.object3D, // THREE.Mesh - used as an overide from the default
                    pointerDownCast: function (data) {
                        var shouldCancel = true;
                        if (!entity.active || entity.cancelFrame) {
                            return;
                        }
                        if (data.castData && pointerId === null) {
                            pointerId = data.id;
                            previousPoint = data.castData.point;

                            // user should only start drawing from within the vicinity of the ball
                            Utils.forEach(players, function (player, i, l, breakLoop) {

                                if (player.getBall() && previousPoint.distanceTo(player.getHandPosition()) < drawDistanceStart) {
                                    previousPoint = player.getHandPosition();
                                    shouldCancel = false;
                                    return;
                                }

                            });

                            if (shouldCancel) {
                                console.log('Did not tap on the ball');
                                pointerId = null;
                                previousPoint = null;
                                EventSystem.fire('missedTap');
                                return;
                            }

                            distanceDrawn = 0;
                            lineArray = [previousPoint];

                            lineComponent.setLine(lineArray);
                            entity.onDrawStart(previousPoint);

                            EventSystem.fire('drawStart');
                        }
                    },
                    pointerMoveCast: function (data) {
                        var point, diff, distance, limiter;
                        if (data.castData && pointerId === data.id && entity.active) {
                            point = data.castData.point.clone();
                            diff = point.clone().sub(previousPoint);
                            distance = previousPoint.distanceTo(point);

                            // check if the distance will exceed max
                            // this prevents the user from moving their fingers very fast to cheat
                            if (distance + distanceDrawn > entity.maxDistance) {
                                // limit draw distance
                                limiter = (entity.maxDistance - distanceDrawn) / distance;
                                diff.multiplyScalar(limiter);
                                point = previousPoint.clone().add(diff);
                            }
                            distanceDrawn += distance;

                            lineArray.push(point);
                            lineComponent.setLine(lineArray);
                            entity.onDraw(lineArray, distanceDrawn, point);
                            lastAngle = previousPoint.angleTo(point);

                            if (lineArray.length > 2) {
                                lastAngle = calculateLastAngle();
                                // var previous = lineArray[lineArray.length - 2];
                                // var current = lineArray[lineArray.length - 1];
                                // lastAngle = Math.atan2(current.y - previous.y, current.x - previous.x);
                            }

                            previousPoint = point;

                            healthbarTarget.position.copy(previousPoint);
                            healthbarTarget.health = (entity.maxDistance - distanceDrawn) / entity.maxDistance * 100;
                            if (entity.maxDistance < 1000) {
                                healthbar.setVisible(!Bento.saveState.load('cheat-hideUi', false));
                            }

                            // did went over the max: force quit drawing
                            if (limiter !== undefined) {
                                // entity.endDrawing(distanceDrawn);
                                entity.active = false;
                            }
                            // debug: show how far is drawn
                            // console.log('Distance drawn: ' + distanceDrawn.toFixed(1));

                            Bento.objects.attach(new Particle({
                                position: previousPoint.clone(),
                                velocity: new THREE.Vector3(
                                    Utils.getRandomRangeFloat(-0.05, 0.05),
                                    Utils.getRandomRangeFloat(0, 0.1),
                                    Utils.getRandomRangeFloat(-0.05, 0.05)
                                )
                            }));
                        }
                    },
                    pointerUpCast: function (data) {
                        if (data.castData && pointerId === data.id) {
                            pointerId = null;
                            healthbar.setVisible(false);
                            entity.onDrawEnd(lineArray, lastAngle, distanceDrawn);
                        }
                    }
                }),
                behavior,

            ]
        }).extend({
            active: true,
            maxDistance: settings.maxDistance || 100,
            cancelFrame: 0,

            // overwrite these
            onDrawStart: function (point) {
                console.log('draw start');
            },
            onDraw: function (array, distance) {},
            onDrawEnd: function (array) {
                console.log('draw end');
            },
            endDrawing: function () {
                // force the drawing to end
                pointerId = null;
                healthbar.visible = false;
                entity.onDrawEnd(lineArray, lastAngle, distanceDrawn);
                healthbar.setVisible(false);
            },
            drawFrom: function (index) {
                lineComponent.visibleFrom(index);
            },
            reset: function () {
                lineComponent.reset();
                distanceDrawn = 0;
                lineArray = [];
                healthbarTarget.health = 100;
            }
        });
        var calculateLastAngle = function () {
            // loop through points, get the average angle over the final 5 meters
            var i = 0;
            var totalDistance = 0;
            var calcDistance = 10;
            var average = 0;
            var current, previous;
            for (i = lineArray.length - 2; i >= 0; --i) {
                current = lineArray[i + 1];
                previous = lineArray[i];
                totalDistance += current.distanceTo(previous);
                if (totalDistance > calcDistance) {
                    current = lineArray[lineArray.length - 1];
                    average = (Math.atan2(current.y - previous.y, current.x - previous.x));
                    break;
                }
            }
            if (totalDistance <= calcDistance) {
                current = lineArray[lineArray.length - 1];
                previous = lineArray[0];
                average = (Math.atan2(current.y - previous.y, current.x - previous.x));
            }

            return average;
        };

        return entity;
    };
    return DrawPlane;
});