/**
 * Game controller, manages game rules
 * @moduleName GameController
 * @snippet GameController|Constructor
GameController({})
 */
bento.define('entities/gamecontroller', [
    'bento',
    'bento/utils',
    'bento/entity',
    'bento/components/anchor',
    'gui/tutorials/drawtutorial',
    'globals',
    'bento/gui/text',
    'bento/math/vector2',
    'bento/components/eventlistener',
    'gui/popups/winpopup',
    'gui/popups/failpopup',
    'bento/components/clickable',
    'bento/math/rectangle',
    'bento/eventsystem',
    'managers/savevalues',
    'modules/watchad',
    'modules/banner',
    'modules/abtest',
    'gui/popups/freegiftpopup',
    'gui/widgets/askrating',
    'random',
    'modules/randomizer',
    'gui/popups/chestpopup',
    'gui/popups/doublecoinspopup',
    'managers/timemanager',
    'gui/popups/dailygiftpopup',
    'modules/notifications',
    'modules/localization'
], function (
    Bento,
    Utils,
    Entity,
    Anchor,
    DrawTutorial,
    Globals,
    Text,
    Vector2,
    EventListener,
    WinPopup,
    FailPopup,
    Clickable,
    Rectangle,
    EventSystem,
    SaveValues,
    WatchAd,
    BannerAd,
    ABTest,
    FreeGiftPopup,
    AskRating,
    Random,
    Randomizer,
    ChestPopup,
    DoubleCoinsPopup,
    TimeManager,
    DailyGiftPopup,
    Notifications,
    Localization
) {
    'use strict';
    var GameController = function (settings) {
        var viewport = Bento.getViewport();
        var isCatchingUp;
        var behavior = new Object({
            name: 'behavior',
            start: function (data) {
                // init tutorial
                if (SaveValues.currentLevel <= 2 && Bento.saveState.load('cheat-skipTutorials', false) === false) {
                    // try to get ball and basket directly
                    Utils.timeout(5, function () {
                        Bento.objects.get('ball', function (ball) {
                            Bento.objects.get('basketHoop', function (hoop) {
                                Bento.objects.attach(new DrawTutorial({
                                    start: ball.position.clone(),
                                    target: hoop.position.clone()
                                }));
                            });
                        });
                    });

                }

                // give daily bonus
                var day = Math.floor(TimeManager.getTime() / 1000 / 60 / 60 / 24);
                if (SaveValues.currentLevel > 2 && SaveValues.dailyBonusKey !== day) {
                    var dailygift = new DailyGiftPopup({
                        dayKey: day,
                        animate: DailyGiftPopup.shouldAnimate,
                        onComplete: function () {
                            if (!Bento.saveState.load('notificationsAllowed', false)) {
                                Notifications.init(function (success) {
                                    Notifications.resetBadgeNumber();
                                    if (success) {
                                        Bento.saveState.save('notificationsAllowed', true);
                                        EventSystem.fire('scheduleNotification', {
                                            id: 'notification',
                                            title: Localization.getText('notifications-title'),
                                            body: Localization.getText('notifications-body' + Utils.getRandom(3)),
                                            time: 24 * 60 * 60 * 1000
                                        });
                                    }
                                });
                            }
                        }
                    });
                    Bento.objects.attach(dailygift);
                }
                DailyGiftPopup.shouldAnimate = true;

                // catching up with tutorial?
                if (SaveValues.currentLevel - 1 > Globals.lvPauseTutorial && Globals.currentPuzzle === Globals.puzzles[Globals.lvPauseTutorial]) {
                    isCatchingUp = true;
                    console.log('Catching up with tutorial');
                }

                if (Bento.saveState.load('cheatTestRating', false)) {
                    askRating(true);
                }

            },
            destroy: function (data) {},
            update: function (data) {},
        });
        var didWin = false;
        var isBonusLevel = SaveValues.levelsUntilBonus <= 0; // cache this until the popups are gone
        var winListener = new EventListener({
            name: 'winListener',
            eventName: 'dunkedBall',
            ignorePause: false,
            onEvent: function (data) {
                var onWinPopup = function (unlockAd) {
                    var checkChest = function (skipInterstitial) {
                        if (SaveValues.keys >= 3) {
                            Bento.objects.attach(new ChestPopup({
                                onComplete: function (keysAd) {
                                    entity.winLevel({
                                        skipInterstitial: skipInterstitial || keysAd
                                    });
                                }
                            }));
                        } else {
                            entity.winLevel({
                                skipInterstitial: skipInterstitial
                            });
                        }
                    };
                    if (isCatchingUp) {
                        // reset
                        Bento.screens.show('screens/main');
                    } else if (SaveValues.currentLevel % 5 === 0) {
                        // show free gift
                        Bento.objects.attach(new FreeGiftPopup({
                            onComplete: function (freeGiftAd) {
                                checkChest(freeGiftAd || unlockAd);
                            }
                        }));
                    } else {
                        checkChest(unlockAd);
                    }
                };
                // show winpopup
                didWin = true;

                // show key
                if (Globals.didGetKey && SaveValues.keys < 3) {
                    SaveValues.keys += 1;
                }

                Utils.timeout(90, function () {
                    if (isBonusLevel) {
                        Bento.objects.attach(new DoubleCoinsPopup({
                            onComplete: function () {
                                // reset right after claiming coins
                                SaveValues.bonusLevelsCleared += 1;
                                SaveValues.levelsUntilBonus = Globals.bonusFrequency * (SaveValues.bonusLevelsCleared % 2 + 1);

                                Bento.objects.attach(new WinPopup({
                                    onComplete: onWinPopup
                                }));
                            }
                        }));
                    } else {
                        Bento.objects.attach(new WinPopup({
                            onComplete: onWinPopup
                        }));
                    }
                });
                EventSystem.fire('GameAnalytics-addDesignEvent', {
                    eventId: "winRate:level-" + SaveValues.currentLevel,
                    value: 1
                });
            }
        });
        var lostTimeout;
        var caughtListener = new EventListener({
            name: 'caughtListener',
            eventName: 'caughtBall',
            ignorePause: false,
            onEvent: function (data) {
                if (lostTimeout) {
                    lostTimeout.stop();
                    lostTimeout.removeSelf();
                    lostTimeout = null;
                }
            }
        });
        var loseListener = new EventListener({
            name: 'loseListener',
            eventName: 'releasedBall',
            ignorePause: false,
            onEvent: function (data) {
                // we wait 3 seconds: if the ball hasn't been dunked at that point we call it a lose
                lostTimeout = Utils.timeout(Bento.saveState.load('cheat-failFast', false) ? 30 : 3 * 60, function () {
                    var winPopup = Bento.objects.get('winPopup');
                    if (!winPopup && !didWin) {
                        // bonus levels: cannot lose
                        if (isBonusLevel) {
                            EventSystem.fire('dunkedBall');
                            return;
                        }

                        Bento.objects.attach(new FailPopup({
                            onComplete: function () {
                                // reset
                                Bento.screens.show('screens/main');

                                if (ABTest.getOption('B110_ISAfterFail') === 'yes') {
                                    showInterstitial();
                                }
                            }
                        }));
                        if (!isCatchingUp) {
                            EventSystem.fire('UserStats-levelFail', {
                                level: SaveValues.currentLevel
                            });
                        }
                        EventSystem.fire('gameOver');
                        EventSystem.fire('GameAnalytics-addDesignEvent', {
                            eventId: "winRate:level-" + SaveValues.currentLevel,
                            value: 0
                        });
                    }
                });
            }
        });
        var entity = new Entity({
            name: 'gameController',
            boundingBox: new Rectangle(0, -64, viewport.width, 128),
            components: [
                behavior,
                winListener,
                caughtListener,
                loseListener,
            ]
        }).extend({
            winLevel: function (data) {
                data = data || {};
                EventSystem.fire('UserStats-levelComplete', {
                    level: SaveValues.currentLevel
                });
                if (!isBonusLevel) {
                    // during bonus the level wasnt played, so we don't increment the current level
                    SaveValues.currentLevel += 1;
                    SaveValues.levelsUntilBonus -= 1;
                }

                Bento.screens.show('screens/main');

                if (!data.skipInterstitial && !entity.skipNextInterstitial) {
                    showInterstitial();
                }
                entity.skipNextInterstitial = false;

                if (SaveValues.currentLevel > 5 && (SaveValues.currentLevel + 6) % 5 === 0) {
                    // ask for rating next level (starts at lv 9)
                    console.log('ask rating next lv');
                    askRating();
                }
            },
            skipNextInterstitial: false,
        });
        var showInterstitial = function () {
            var skipAfterAB = ABTest.getOption('B101_StartIS');
            var skipAfter = 4;
            switch (skipAfterAB) {
            case 'After2':
                skipAfter = 3;
                break;
            case 'After3':
                skipAfter = 4;
                break;
            case 'After4':
                skipAfter = 5;
                break;
            default:
                skipAfter = 4;
            }
            if (SaveValues.currentLevel >= skipAfter && !Bento.saveState.load('cheat-skipInterstitials', false) && !Bento.saveState.load('noAds', false)) {
                EventSystem.fire('GameAnalytics-addDesignEvent', {
                    eventId: 'Ads:is:interstitialStart'
                });
                WatchAd.showInterstitial({
                    tag: '',
                    onComplete: function (error) {
                        if (error) {
                            console.log('WARNING: interstitial failed to show ' + error);
                        }
                        EventSystem.fire('GameAnalytics-addDesignEvent', {
                            eventId: 'Ads:is:interstitialComplete'
                        });
                    }
                });
            }
        };
        var askRating = function (force) {
//            if (AskRating.canAsk(force)) {
//                if (Utils.isNativeIos()) {
//                    AskRating.askIos10();
//                } else {
//                    Bento.objects.attach(new AskRating({
//                        z: Globals.Layers.POPUPS + 1,
//                        force: force
//                    }));
//                    if (force) {
//                        Bento.saveState.save('cheatTestRating', false);
//                    }
//                }
//            }
        };

        return entity;
    };
    var excluded = ['level-1', 'level-2', 'level-17-2p', 'pause-tutorial'];
    GameController.getPuzzle = function () {
        var level = SaveValues.currentLevel - 1; // 1-based index
        var puzzleCount = Globals.puzzles.length;
        // var randomizer = new Random(level); // randomizer not working as expected
        var randomizer = new Randomizer({
            seed: level
        });
        var puzzleIndex = 0;
        var puzzle = '';

        if (SaveValues.levelsUntilBonus <= 0) {
            return Globals.bonusLevels[Utils.modulo(Math.floor(SaveValues.currentLevel / 10 - 1), Globals.bonusLevels.length)];
        }

        // special case: user didn't get the pause tutorial yet due to old version
        if (SaveValues.currentLevel > Globals.lvPauseTutorial && Bento.saveState.load('pauseTutorial', false) === false) {
            return Globals.puzzles[Globals.lvPauseTutorial];
        }

        if (level < puzzleCount) {
            return Globals.puzzles[level % puzzleCount];
        } else {
            // console.log('randomized level', level, Math.floor(new Random(level).getRandom() * puzzleCount));
            // randomize levels so user is less likely to notice the repetition
            while (!puzzle || excluded.indexOf(puzzle) >= 0) {
                // puzzleIndex = Math.floor(randomizer.getRandom() * puzzleCount);
                puzzleIndex = Math.floor(randomizer.getRandom(puzzleCount));
                puzzle = Globals.puzzles[puzzleIndex];
            }
            console.log('Randomized level');
            return puzzle;
        }

    };

    if (Bento.isDev()) {
        window.win = function () {
            EventSystem.fire('dunkedBall');
        };
    }
    return GameController;
});