/**
 * Platform
 * @moduleName Platform
 */
bento.define('entities/platform', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'modules/box2d',
    'globals',
    'onigiri/onigiri',
    'onigiri/entity3d',
    'components/pathfollower'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Box2D,
    Globals,
    Onigiri,
    Entity3D,
    PathFollower
) {
    'use strict';
    return function (settings) {
        var position = settings.position.clone() || new Vector2(0, 0);
        var path = settings.polygon || [new Vector2(0, 0), new Vector2(settings.dimension.width || 16, 0), new Vector2(settings.dimension.width || 16, settings.dimension.height || 16), new Vector2(0, settings.dimension.height || 16)];
        var levelScale = settings.levelScale || 1;
        var depth = settings.depth || 1;
        var flipOffset = settings.flipOffset || 0;
        var createMesh = function (vertices) {
            var shape = new THREE.Shape();
            Utils.forEach(vertices, function (vertex, i, l, breakLoop) {
                if (i === 0) {
                    shape.moveTo(vertex[0], vertex[1]);
                }
                shape.lineTo(vertex[0], vertex[1]);

            });
            var extrudeSettings = {
                steps: 8,
                depth: depth * levelScale,
                bevelEnabled: true,
                bevelThickness: 8 * Onigiri.pxToThree,
                bevelSize: 6 * Onigiri.pxToThree,
                bevelOffset: -4 * Onigiri.pxToThree,
                bevelSegments: 7
            };
            var outlineExtrudeSettings = Utils.cloneJson(extrudeSettings);
            outlineExtrudeSettings.bevelSize = 8 * Onigiri.pxToThree;

            var geometry = new THREE.ExtrudeGeometry(shape, extrudeSettings);
            var material = new THREE.MeshPhongMaterial({
                color: Globals.getWallColor(),
                // specular: 0xfefeb6,
                specular: 0xffffff,
                shininess: 4,
                reflectivity: 0.25,
            });

            var outlineGeometry = new THREE.ExtrudeGeometry(shape, outlineExtrudeSettings);
            var outlineMaterial = new THREE.MeshBasicMaterial({
                color: 0x010b13,
                side: THREE.BackSide,
                opacity: 0.5,
                transparent: true
            });

            var mesh = new THREE.Mesh(geometry, material);
            var outline = new THREE.Mesh(outlineGeometry, outlineMaterial);
            mesh.add(outline);
            return mesh;
        };

        var decomp = window.decomp;
        var concavePolys = (function (p) {
            var c = p.map(function (v) {
                return [
                    v.x * Onigiri.pxToThree * levelScale,
                    v.y * -1 * Onigiri.pxToThree * levelScale + flipOffset
                ];
            });
            decomp.makeCCW(c);
            return c;
        })(path);

        var createFixtures = function (p) {
            var convexPolys = decomp.quickDecomp(p);

            convexPolys.forEach(function (poly) {
                decomp.removeCollinearPoints(poly);
            });
            var fixtures = convexPolys.map(function (poly) {
                var vertices = poly.map(function (v) {
                    return new Vector2(v[0], v[1]);
                });

                return {
                    shape: new Box2D.PolygonShape(vertices),
                    // filter,
                    // friction: friction,
                    friction: Globals.wallFriction,
                    restitution: Globals.wallRestitution,
                    // density,
                    // onBeginContact,
                    // etc.
                };
            });
            return fixtures;
            // now you can pass fixtures to new Box2D.Body
        };

        // var box2dPath = fixConcavePath(copyPath(path));
        var body = new Box2D.Body({
            type: 'kinematic',
            fixture: createFixtures(concavePolys)
        });

        var platformMesh = createMesh(concavePolys);

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                if (settings.properties && settings.properties.follow) {
                    var follow = new PathFollower({
                        pathName: settings.properties.follow,
                    });
                    entity.attach(follow);
                }
            },
            destroy: function (data) {},
            update: function (data) {},
            draw: function (data) {}
        };
        var entity = new Entity3D({
            z: 2,
            object3D: platformMesh,
            name: 'platform' + settings.properties.name,
            position: new THREE.Vector3(position.x * Onigiri.pxToThree * levelScale, position.y * Onigiri.pxToThree * levelScale, position.z * Onigiri.pxToThree),
            components: [
                body,
                behavior,
            ]
        });
        return entity;
    };
});