/**
 * Player basketballer
 * @moduleName Player
 * @snippet Player()|Constructor
Player({})
 */
bento.define('entities/player', [
    'bento',
    'bento/utils',
    'onigiri/entity3d',
    'onigiri/onigiri',
    'components/materialswapper',
    'entities/basketball',
    'onigiri/animator',
    'bento/components/eventlistener',
    'globals',
    'modules/box2d',
    'bento/math/vector2',
    'bento/eventsystem',
    'managers/savevalues',
    'gui/widgets/taphand'
], function (
    Bento,
    Utils,
    Entity3D,
    Onigiri,
    MaterialSwapper,
    BasketBall,
    Animator,
    EventListener,
    Globals,
    Box2D,
    Vector2,
    EventSystem,
    SaveValues,
    TapHand
) {
    'use strict';
    var PlayerGraphic = function (settings) {
        var direction = settings.direction;
        var solidColor = Bento.saveState.load('cheat-playerskin', 'default') === 'green/red';
        var mesh = Onigiri.getMesh('players/baller-' + SaveValues.currentPlayer + '.gltf');
        var animator = new Animator({
            object3D: Onigiri.getMesh('players/baller-default.gltf'), // get the animations from the default
            rootObject: mesh,
            defaultAnimation: 'hold-' + (direction === 1 ? 'faceright' : 'faceleft'),
            onFinish: function (evt) {
                var name = evt.action.getClip().name;
                if (name.startsWith('shoot')) {
                    animator.stop();
                    animator.play('emptyhanded-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                        loop: true
                    }, false);
                }
            }
        });
        var meshContainer = new Entity3D({
            name: 'meshContainer',
            object3D: mesh, // THREE.object3D reference
            castShadow: true,
            components: [
                Globals.getPlayerMaterial(SaveValues.currentPlayer, null, solidColor),
                animator
            ]
        });
        var ballBone = (function () {
            var item;
            mesh.traverse(function (child) {
                if (child.name === 'item') {
                    item = child;
                }
            });
            return item || Utils.getSafe(mesh, ['children', 0, 'children', 0], null);
        })();

        meshContainer.extend({
            animator: animator,
            ballBone: ballBone
        });

        return meshContainer;
    };
    var Player = function (settings) {
        var position = settings.position;
        var direction = settings.direction || 1;
        var noBall = settings.noBall || false;
        var ball = noBall ? null : new BasketBall({});
        // var ball = null;
        var ballOffset = new THREE.Vector3();
        var threwBall = false;
        var behavior = new Object({
            name: 'behavior',
            start: function (data) {
                if (ball) {
                    ball.player = entity;
                    if (!ball.isAdded) {
                        Bento.objects.attach(ball);
                    }
                    setBallInHand();
                    Bento.objects.draw();
                    ball.snapTrail();
                } else {
                    meshContainer.animator.stop();
                    meshContainer.animator.play('emptyhanded-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                        loop: true
                    }, false);
                }

                if (direction === -1) {
                    entity.euler.y += Math.PI;
                }

            },
            destroy: function (data) {},
            update: function (data) {
                setBallInHand();
            },
            draw: function (data) {}
        });
        var meshContainer = new PlayerGraphic({
            direction: direction
        });

        var width = 1.5;
        var height = 6;
        var body = new Box2D.Body({
            type: 'static',
            fixedRotation: true,
            fixture: {
                // BoxShape is halfW, halfH
                isSensor: true,
                shape: new Box2D.BoxShape(width / 2, height / 2, new Vector2(0, height / 2)),
                friction: 1,
                restitution: 0,
                density: 10000,
                onBeginContact: function (contact) {
                    if (contact && contact.entity && contact.entity.name === 'ball' && !ball && !ignoreBall) {
                        ball = contact.entity;
                        if (ball.player === entity) {
                            // cannot be caught by the player who threw it
                            ball = null;
                            return;
                        }
                        ball.setKinematic(function () {
                            setBallInHand();
                        });
                        EventSystem.fire('caughtBall');
                        // console.log('begin', contact);
                    }
                },
            }
        });
        var ignoreBall = false;
        var winListener = new EventListener({
            name: 'winListener',
            eventName: 'dunkedBall',
            ignorePause: false,
            onEvent: function (data) {
                meshContainer.animator.stop();
                meshContainer.animator.play('celebrate-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                    loop: false
                }, false);

                // disable sensor
                ignoreBall = true;
            }
        });
        var loseListener = new EventListener({
            name: 'loseListener',
            eventName: 'gameOver',
            ignorePause: false,
            onEvent: function (data) {
                if (ball) {
                    return;
                }
                meshContainer.animator.stop();
                meshContainer.animator.play('miss-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                    loop: false
                }, false);
                entity.remove(winListener);

                // disable sensor
                ignoreBall = true;
            }
        });
        var skinListener = new EventListener({
            name: 'skinListener',
            eventName: 'refreshPlayerSkin',
            ignorePause: true,
            onEvent: function (data) {
                meshContainer.removeSelf();
                meshContainer = new PlayerGraphic({
                    direction: direction
                });
                entity.attach(meshContainer);
            }
        });
        var didShowHand = false;
        var tapHand;
        var tutorialListener = new EventListener({
            name: 'tutorialListener',
            eventName: 'missedTap',
            ignorePause: false,
            onEvent: function (data) {
                if (ball && ball.player === entity && !didShowHand && !Bento.saveState.load('cheat-hideUi', false)) {
                    didShowHand = true;
                    tapHand = new TapHand({
                        z: Globals.Layers.GUI - 1,
                        position: Onigiri.projectToBento(entity.getHandPosition()),
                        rotation: direction === 1 ? Math.PI / 8 : -Math.PI / 8,
                        imageName: 'gui/hand',
                        upPosition: new Vector2(0, -40),
                        originRelative: new Vector2(0, 0),
                        handSize: 48,
                        speed: 1,
                        startInUpPosition: false,
                        onTap: function () {}
                    });
                    Bento.objects.attach(tapHand);
                }
            }
        });
        var tutorialEndListener = new EventListener({
            name: 'tutorialListener',
            eventName: 'drawStart',
            ignorePause: false,
            onEvent: function (data) {
                if (didShowHand && tapHand) {
                    tapHand.removeSelf();
                }
            }
        });
        var entity = new Entity3D({
            name: 'player',
            // receiveShadow: true,
            position: position,
            scale: new THREE.Vector3(2.4, 2.4, 2.4),
            euler: new THREE.Euler(0, Math.PI / 2, 0),
            family: ['players', 'ballTargets'],
            components: [
                meshContainer,
                behavior,
                winListener,
                loseListener,
                skinListener,
                tutorialListener,
                tutorialEndListener,
                body
            ]
        }).extend({
            getHandPosition: function () {
                return ballOffset;
            },
            shoot: function () {
                meshContainer.animator.stop();
                meshContainer.animator.play('shoot-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                    loop: false
                }, false);

                // disconnect ball
                threwBall = true;
                ball = null;
            },
            getBall: function () {
                return ball;
            }
        });
        var setBallInHand = function () {
            // set ball offset to animation
            var rotation = new THREE.Quaternion();
            var scale = new THREE.Vector3();
            if (meshContainer.ballBone && ball && !ignoreBall) {
                ball.player = entity;
                meshContainer.ballBone.matrixWorld.decompose(ballOffset, rotation, scale);
                if (ball) {
                    ball.position.set(
                        ballOffset.x + 0.75 * direction,
                        ballOffset.y,
                        ballOffset.z // 0?
                    );
                    // ball.position.copy(ballOffset);
                    ball.quaternion.copy(rotation);

                    var ballSettings = Globals.ballSettings[SaveValues.currentBall] || {};
                    ball.euler.x += ballSettings.eulerX || 0;
                    ball.euler.y += ballSettings.eulerY || 0;
                    ball.euler.z += ballSettings.eulerZ || 0;
                    if (!ball.isAdded) {
                        Bento.objects.attach(ball);
                    }
                }
                meshContainer.animator.play('hold-' + (direction === 1 ? 'faceright' : 'faceleft'), {
                    loop: true
                }, false);
            }
        };

        return entity;
    };
    return Player;
});