/**
 * Button with an icon on top
 * @moduleName IconButton
 * @snippet IconButton|snippet
IconButton({
    z: ${1:0},
    position: ${2:new Vector2(0, 0)},
    imageName: '${3:gui/buttons/wide_yellow}',
    frameCountX: ${4:1},
    frameCountY: ${5:3},

    // icon options
    icon: new Entity({}),
    iconSprite: new Sprite({}),
    iconName: '',
    
    // offset of icon
    offset: new Vector2(0, 0),
    offsetWhenPressed: 0, // push pixels down when onButtonDown

    // icon transparency when inactive
    iconAlpha: 0.5,
    
    onClick: function () {
        $6
    }
});
 */
bento.define('gui/iconbutton', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var entitySettings = {};
        var entity;
        var iconEntity = settings.icon;
        var iconSprite = settings.iconSprite;
        var iconAlpha = settings.iconAlpha;
        var iconName = settings.iconName || '';
        var offset = settings.offset || new Vector2(0, 0);
        var offsetWhenPressed = settings.offsetWhenPressed || 0;
        var icon = iconEntity || new Entity({
            name: 'icon',
            components: [
                iconSprite || new Sprite({
                    imageName: iconName,
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });

        // setup icon position
        icon.position = offset.clone();

        // defaults
        entitySettings.frameCountX = entitySettings.frameCountX || 1;
        entitySettings.frameCountY = entitySettings.frameCountY || 3;

        // take over properties
        Utils.extend(entitySettings, settings);
        
        // overwrite callbacks
        entitySettings.onButtonDown = function () {
            icon.position.y = offset.y + offsetWhenPressed;
            if (settings.onButtonDown) {
                settings.onButtonDown.apply(entity);
            }
        };
        entitySettings.onButtonUp = function () {
            icon.position.y = offset.y;
            if (settings.onButtonUp) {
                settings.onButtonUp.apply(entity);
            }
        };

        // merge components array
        entitySettings.components = [
            icon
        ].concat(settings.components || []);

        // spawn entity
        entity = new ClickButton(entitySettings);

        // allow to change icon
        entity.extend({
            changeIcon: function (newIcon, offsetPosition) {
                icon.getComponent('sprite', function (sprite) {
                    sprite.setup(newIcon);
                    if (offsetPosition) {
                        icon.position = offsetPosition.clone();
                    }
                });
            },
            icon: icon
        });

        if (iconAlpha) {
            // add behavior to set alpha of icon when button is disabled
            entity.attach({
                name: 'alpha',
                update: function () {
                    if (entity.isActive()) {
                        icon.alpha = 1;
                    } else {
                        icon.alpha = iconAlpha;
                    }
                }
            });
        }

        return entity;
    };
});