/**
 * Overlay to show what the game will look like on an iPhone X, or potentially other 'notch' devices.
 * Attach this entity to the game once at the end of the preloader. It will persist between screens.
 * The entity will only be visible if the viewport aspect ratio matches that of the iPhone X.
 *
 * Settings:
 * `landscapeMode` - When the viewport is wide, this determines if the overlay is rotated clockwise ('right') or counterclockwise ('left').
 * `overrideInsets` - The overlay will affect the screen's safe zone, influencing any Anchor components that are used in the game.
 *
 * @moduleName IPhoneXOverlay
 * @snippet IPhoneXOverlay.snippet
IPhoneXOverlay({
    z: Globals.Layers.IPHONE_X_OVERLAY,
    overrideInsets: true,
    landscapeMode: 'left',
})
 */
bento.define('gui/iphonexoverlay', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/components/anchor'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    Anchor
) {
    'use strict';

    var ratioLandscape = 2436 / 1125;
    var ratioPortrait =  1125 / 2436;
    var isIPhoneXAspectRatio = function (viewport) {
        var ratio = viewport.width / viewport.height;
        return Math.abs(ratio - ratioLandscape) < 0.01
            || Math.abs(ratio - ratioPortrait)  < 0.01;
    };

    return function (settings) {
        var viewport = Bento.getViewport();
        
        var overrideInsets = Utils.getDefault(settings.overrideInsets, true);
        var landscapeMode = Utils.getDefault(settings.landscapeMode, 'left');

        var sprite = new Sprite({ 
            imageName: 'iphone-x-overlay/iphone-x',
            originRelative: new Vector2(0.5, 0.5)
        });
        var menuBar = new Entity({
            name: 'menuBar',
            position: new Vector2(0, sprite.frameHeight * 0.485),
            components: [
                new Sprite({ 
                    imageName: 'iphone-x-overlay/iphone-x-menu-bar',
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });

        var onResize = function () {

            var isLandscape = (viewport.width > viewport.height);

            if (isLandscape) {
                entity.rotation = Math.PI * 0.5 * (landscapeMode === 'left' ? -1 : 1);
                entity.position.x = viewport.width * 0.5;
                entity.position.y = viewport.height * 0.5;
                entity.scale.x = (viewport.height+1) / entity.dimension.width;
                entity.scale.y = (viewport.width+1) / entity.dimension.height;
                menuBar.rotation = Math.PI * 0.5;
                menuBar.position.x = sprite.frameWidth * 0.47 * (landscapeMode === 'left' ? -1 : 1);
                menuBar.position.y = 0;
                menuBar.scale.x = 1.52;
            } else {
                entity.rotation = 0;
                entity.position.x = viewport.width * 0.5;
                entity.position.y = viewport.height * 0.5;
                entity.scale.x = (viewport.width+1) / entity.dimension.width;
                entity.scale.y = (viewport.height+1) / entity.dimension.height;
                menuBar.rotation = 0;
                menuBar.position.x = 0;
                menuBar.position.y = sprite.frameHeight * 0.485;
                menuBar.scale.x = 1.0;
            }

            if (isIPhoneXAspectRatio(viewport)) {
                entity.visible = true;
                if (overrideInsets) {
                    if (isLandscape) {
                        Anchor.insetOverrides.top = null;
                        Anchor.insetOverrides.bottom = viewport.height * 0.056 + 50; // extra banner
                        Anchor.insetOverrides.left = viewport.width * 0.054;
                        Anchor.insetOverrides.right = viewport.width * 0.054;
                    } else {
                        Anchor.insetOverrides.left = null;
                        Anchor.insetOverrides.right = null;
                        Anchor.insetOverrides.top = viewport.height * 0.052;
                        Anchor.insetOverrides.bottom = viewport.height * 0.042 + 50; // extra banner
                    }
                }
            } else {
                entity.visible = false;
                if (overrideInsets) {
                    Anchor.insetOverrides.top = null;
                    Anchor.insetOverrides.bottom = 50; // extra banner
                    Anchor.insetOverrides.left = null;
                    Anchor.insetOverrides.right = null;
                }
            }
        };

        var behavior = {
            name: 'behavior',
            start: function () {
                onResize();
                EventSystem.on('resize', onResize);
            },
            destroy: function () {
                EventSystem.off('resize', onResize);
            },
        };

        var entity = new Entity({
            z: settings.z,
            name: 'iPhoneXOverlay',
            float: true,
            global: true,
            position: new Vector2(viewport.width*0.5, viewport.height*0.5),
            components: [sprite, behavior, menuBar],
        });

        if (entity.z === 0) {
            Utils.log('WARNING: iPhoneXOverlay has a low Z index.');
        }

        return entity;
    };
});