/**
 * Asks for a rating once
 * @moduleName AskRating
 */
bento.define('gui/widgets/askrating', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/localization',
    'gui/dialogs/basedialog',
    'gui/hdbuttonsprite',
    'gui/iconbutton'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    ClickButton,
    Counter,
    Text,
    EventSystem,
    Utils,
    Tween,
    Localization,
    Dialog,
    HdButtonSprite,
    IconButton
) {
    'use strict';
    var AskRating = function (settings) {
        var dialog;
        if (!AskRating.canAsk(settings.force)) {
            if (settings.onComplete) {
                settings.onComplete();
            }
            // return a dummy
            return new Entity({});
        }

        dialog = new Dialog({
            z: settings.z,
            name: 'askRating',
            title: '',
            text: Localization.getText('askLike').replace('{TITLE}', Localization.getText('title')),
            onComplete: function () {
                Bento.saveState.save('askedRatingVersion', window.VERSION);
                if (settings.onComplete) {
                    settings.onComplete();
                }
            }
        });

        var yesText = new Text({
            position: new Vector2(0, 0),
            text: Localization.getText('yesRatings'),
            font: 'omnes',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            strokeStyle: '#11ce4f',
            ySpacing: 4,
            lineWidth: 2.5,
            maxWidth: 80,
            maxHeight: 64,
            linebreaks: false,
            drawDebug: false,
            components: []
        });
        var yesButton = new IconButton({
            name: 'yesButton',
            position: new Vector2(36, 64),
            sprite: HdButtonSprite({
                imageName: 'gui/buttons/green',
                width: 64,
                height: null
            }),
            icon: yesText,
            offset: new Vector2(-3, -1.5),
            offsetWhenPressed: 2, // push pixels down when onButtonDown
            onClick: function () {
                // redirect to app store
                var link = Utils.getStoreLink();
                if (Utils.isNativeIos() || Utils.isDev()) {
                    link += '?action=write-review';
                }
                Utils.openUrl(link);
                analytics(true);
                dialog.dismiss();
            }
        });

        var noText = new Text({
            position: new Vector2(0, 0),
            text: Localization.getText('noRatings'),
            font: 'omnes',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            strokeStyle: '#ed2414',
            ySpacing: 4,
            lineWidth: 2.5,
            maxWidth: 80,
            maxHeight: 64,
            linebreaks: false,
            drawDebug: false,
            components: []
        });
        var noButton = new IconButton({
            name: 'noButton',
            position: new Vector2(-36, 64),
            sprite: HdButtonSprite({
                imageName: 'gui/buttons/red',
                width: 64,
                height: null
            }),
            icon: noText,
            offset: new Vector2(-3, -1.5),
            offsetWhenPressed: 2, // push pixels down when onButtonDown
            onClick: function () {
                // redirect to support doc
                var link = 'https://docs.google.com/forms/d/e/1FAIpQLSdtODSDEnRWXQDR-_YxsBsz9d25YOKE8rCFwoVj27vu1qFrxA/viewform';
                Utils.openUrl(link);
                analytics(false);
                dialog.dismiss();
            }
        });

        dialog.attach(yesButton);
        dialog.attach(noButton);

        dialog.scale.x = 2;
        dialog.scale.y = 2;

        dialog.attach(new Object({
            name: 'behavior',
            start: function (data) {
                new Tween({
                    from: 0,
                    to: dialog.scale.x,
                    in: 30,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 2,
                    onStart: function () {},
                    onUpdate: function (v, t) {
                        dialog.scale.x = v;
                        dialog.scale.y = v;
                    },
                    onComplete: function () {

                    }
                });
            },
        }));

        return dialog;
    };

    var analytics = function (pressedToRate) {
        EventSystem.fire('GameAnalytics-addDesignEvent', {
            id: "Stats:RatingPopup:ratingShown"
        });

        if (pressedToRate) {
            EventSystem.fire('GameAnalytics-addDesignEvent', {
                id: "Stats:RatingPopup:ratedUs"
            });
        } else {
            EventSystem.fire('GameAnalytics-addDesignEvent', {
                id: "Stats:RatingPopup:pressedNo"
            });
        }

        EventSystem.fire("FirebaseAnalytics-logEvent", {
            name: "ratingPopup",
            properties: {
                name: 'ratingShown',
                action: pressedToRate ? 'ratedUs' : 'pressedNo'
            }
        });
    };

    AskRating.canAsk = function (force) {
        // return true;
        if (Bento.saveState.load('askedRatingVersion', false) === window.VERSION) {
            // already asked before
            return force || false;
        } else {
            return true;
        }
    };

    AskRating.askIos10 = function (settings) {
        var parseVersion = function (versionString) {
            var versions = versionString.split('.');
            var versionArray = [];
            Utils.forEach(versions, function (numberStr) {
                versionArray.push(parseInt(numberStr));
            });
            return versionArray;
        };

        if (!AskRating.canAsk()) {
            return;
        }

        var version = [0, 0, 0];
        if (window.device) {
            version = parseVersion(window.device.version);
        }

        // set rating in motion so should save this
        Bento.saveState.save('askedRatingVersion', window.VERSION);

        // from iOS 10.3.0 there is a new way to present reviews
        if (version[0] * 10000 + version[1] * 100 + version[0] >= 100300) {
            // use the new skstorereviewcontroller, this may or may not appear at all
            try {
                if (window.inappreview) {
                    analytics(true);
                    window.inappreview.requestReview(function () {
                        // note: these callbacks are not programmed at all in the plugin...
                    }, function () {
                        // Utils.log('ERROR: inappreview plugin failed');
                    });
                } else {
                    Utils.log("ERROR: inappreview plugin missing");
                }
            } catch (e) {
                Utils.log("ERROR: " + e);
            }
        } else {
            // don't do anything, asking for review is against app store policy (?)
        }

    };

    return AskRating;
});