/**
 * Fountain/object rain effect
 * @moduleName Fountain
 * @snippet Fountain|constructor
Fountain({})
 */
bento.define('gui/widgets/fountain', [
    'bento',
    'bento/utils',
    'bento/math/vector2',
    'bento/tween',
    'bento/entity',
    'bento/math/rectangle',
    'globals',
    'bento/gui/text',
    'bento/components/anchor',
    'bento/components/spritecontainer'
], function (
    Bento,
    Utils,
    Vector2,
    Tween,
    Entity,
    Rectangle,
    Globals,
    Text,
    Anchor,
    SpriteContainer
) {
    'use strict';
    var Coin = function (settings) {
        var coinCounter = settings.coinCounter;
        var target = settings.target;
        var scale = settings.scale || 1;
        var onDestroy = settings.onDestroy;
        var sprite = new SpriteContainer({
            imageName: settings.imageName,
            height: settings.size
        });
        var gravity = settings.gravity || 0.3;
        var rotationSpeed = settings.rotationSpeed || 0;
        var behavior = {
            name: 'behavior',
            gravity: gravity,
            velocity: settings.velocity || new Vector2(0, 0),
            update: function (data) {
                behavior.velocity.y += behavior.gravity;
                entity.position.x += behavior.velocity.x;
                entity.position.y += behavior.velocity.y;

                if (rotationSpeed) {
                    entity.rotation += rotationSpeed;
                }
            },
            start: function () {
                // after a few frames of flying, fly towards coin counter in topbar
                new Tween({
                    from: 1,
                    to: 0,
                    in: 45,
                    delay: settings.delay || 30,
                    ease: 'cubic',
                    alpha: 5,
                    onUpdate: function (v, t) {
                        behavior.gravity = gravity * v;
                        behavior.velocity.x = behavior.velocity.x * v;
                        behavior.velocity.y = behavior.velocity.y * v;
                        entity.position.x += (target.x - entity.position.x) * (1 - v);
                        entity.position.y += (target.y - entity.position.y) * (1 - v);

                        entity.scale.x = (v + 0.5) / 1.5 * scale;
                        entity.scale.y = (v + 0.5) / 1.5 * scale;
                    },
                    onComplete: function () {
                        entity.removeSelf();
                        // Bento.audio.stopSound('sfx_coinpickup');
                        // Bento.audio.playSound('sfx_coinpickup');

                        if (coinCounter) {
                            new Tween({
                                from: 1.25,
                                to: 1,
                                in: 20,
                                ease: 'elastic',
                                decay: 5,
                                oscillations: 1,
                                onUpdate: function (v, t) {
                                    coinCounter.scale.x = v;
                                    coinCounter.scale.y = v;
                                },
                                onComplete: function () {
                                    if (onDestroy) {
                                        onDestroy();
                                    }
                                }
                            });
                        } else {
                            if (onDestroy) {
                                onDestroy();
                            }
                        }
                    }
                });
            }
        };
        var entity = new Entity({
            z: settings.z,
            name: 'coin',
            position: settings.position,
            components: [
                sprite,
                behavior
            ]
        });

        return entity;
    };
    var CoinRain = function (settings) {
        var safeZone = Anchor.getSafeZone();
        var type = settings.type || 'coins';
        var imageName = Utils.getDefault(settings.imageName, 'gui/icons/big-coin');
        var spriteSize = settings.spriteSize || 16;
        var onComplete = settings.onComplete;
        var onDestroy = settings.onDestroy;
        var onCoin = settings.onCoin;
        var targetPosition = settings.targetPosition || new Vector2(0, 0);
        var spawnPosition = settings.spawnPosition || new Vector2(safeZone.width / 2, safeZone.height / 2);
        var offsetRandomization = settings.randomization || 0;
        var extraAmount = settings.extraAmount || false;
        var velocityScale = settings.velocityScale || 1;
        var gravity = settings.gravity;
        var coinPercentage = 0;
        var coinTotal = settings.coinTotal || (extraAmount ? 4 : 2) * 10;
        var rainBehavior = new Object({
            name: 'rainBehavior',
            update: function (data) {
                // rain some coins
                var coin;
                if (entity.ticker % 1 === 0 && entity.ticker < coinTotal) {
                    coin = new Coin({
                        imageName: imageName,
                        size: spriteSize,
                        type: type,
                        gravity: gravity,
                        rotationSpeed: type === 'coins' ? 0 : Utils.getRandomRangeFloat(-Math.PI / 20, Math.PI / 20),
                        target: targetPosition,
                        position: spawnPosition.add(new Vector2(
                            Utils.getRandomRangeFloat(-offsetRandomization, offsetRandomization),
                            Utils.getRandomRangeFloat(-offsetRandomization, offsetRandomization)
                        )),
                        velocity: new Vector2(
                            Utils.getRandomRangeFloat(-3, 3) * velocityScale,
                            Utils.getRandomRangeFloat(-8, -2) * velocityScale
                        ),
                        onDestroy: function () {
                            if (onDestroy) {
                                onDestroy(coin, coinPercentage);
                            }
                            coinPercentage += 1 / coinTotal;
                            behavior.coins -= 1;
                        }
                    });
                    if (onCoin) {
                        onCoin(coin);
                    }
                    behavior.coins += 1;
                    behavior.didStart = true;
                    coinRainContainer.attach(coin);
                }
            },
        });
        var coinRainContainer = new Entity({
            name: 'coinContainer',
            components: [
                rainBehavior
            ]
        });
        var behavior = new Object({
            name: 'behavior',
            coins: 0,
            didStart: false,
            start: function (data) {},
            destroy: function (data) {
                coinRainContainer.removeSelf();
            },
            update: function () {
                if (behavior.didStart && behavior.coins <= 0) {
                    entity.removeSelf();
                    if (onComplete) {
                        onComplete();
                    }
                }
            }
        });
        var entity = new Entity({
            z: settings.z || Globals.Layers.GUI,
            name: 'coinRain',
            updateWhenPaused: Bento.objects.isPaused(),
            components: [
                behavior,
                coinRainContainer
            ]
        });
        return entity;
    };
    CoinRain.Coin = Coin;
    return CoinRain;
});