/**
 * Tutorial hand - shows a tapping gesture
 * @moduleName TapHand
 * @snippet TapHand()|constructor
TapHand({
    z: 0,
    position: new Vector(0, 0),
    rotation: 0,
    imageName: 'gui/taphand',
    upPosition: new Vector2(0, -30),
    speed: 1,
    startInUpPosition: false,
    onTap: function () {}
})
 */
bento.define('gui/widgets/taphand', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'bento/components/spritecontainer'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    SpriteContainer
) {
    'use strict';
    return function (settings) {
        var tapPosition = new Vector2(0, 0);
        var upPosition = settings.upPosition || new Vector2(0, -30);
        var speed = settings.speed || 1;
        var rotation = settings.rotation || 0;
        var startInUpPosition = settings.startInUpPosition || false;
        var imageName = settings.imageName || 'gui/hand';
        var onTap = settings.onTap || function () {};

        var sprite = new SpriteContainer({
            width: settings.handSize || 64,
            imageName: imageName,
            originRelative: settings.originRelative || new Vector2(0.5, 0)
        });

        var moveUp = function () {
            var completed = false;
            new Tween({
                ignoreGameSpeed: true,
                from: 0,
                to: upPosition.y,
                in: Math.floor(24 / speed),
                delay: Math.floor(12 / speed),
                ease: Tween.EASEINOUTQUAD,
                onUpdate: function (v, t) {
                    entity.position.y = tapPosition.y + v;
                },
                onComplete: function () {
                    if (completed) {
                        return;
                    }
                    completed = true;
                    moveDown();
                }
            });

            new Tween({
                ignoreGameSpeed: true,
                from: 0,
                to: upPosition.x,
                in: Math.floor(24 / speed),
                delay: Math.floor(12 / speed),
                ease: Tween.EASEINOUTQUAD,
                onUpdate: function (v, t) {
                    entity.position.x = tapPosition.x + v;
                }
            });
        };

        var moveDown = function () {
            var completed = false;
            new Tween({
                ignoreGameSpeed: true,
                from: upPosition.y,
                to: 0,
                in: Math.floor(8 / speed),
                delay: Math.floor(18 / speed),
                ease: Tween.EASEINQUAD,
                onUpdate: function (v, t) {
                    entity.position.y = tapPosition.y + v;
                },
                onComplete: function () {
                    if (onTap) {
                        onTap();
                    }
                    if (completed) {
                        return;
                    }
                    completed = true;
                    moveUp();

                    new Tween({
                        ignoreGameSpeed: true,
                        from: -0.2,
                        to: 0,
                        in: 16,
                        ease: Tween.EASEINOUTBACK,
                        onUpdate: function (v, t) {
                            entity.scale.y = 1 + v;
                            entity.scale.x = 1 - v;
                        },
                        onComplete: function () {}
                    });
                }
            });

            new Tween({
                ignoreGameSpeed: true,
                from: upPosition.x,
                to: 0,
                in: Math.floor(8 / speed),
                delay: Math.floor(18 / speed),
                ease: Tween.EASEINQUAD,
                onUpdate: function (v, t) {
                    entity.position.x = tapPosition.x + v;
                }
            });
        };
        var hasStarted = false;
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                sprite.scale.y *= -1;
                sprite.rotation = -Math.PI / 10;
                if (hasStarted) {
                    return;
                }
                hasStarted = true;
                if (!startInUpPosition) {
                    moveUp();
                } else {
                    entity.position = upPosition.clone();
                    moveDown();
                }
            }
        };
        var entity = new Entity({
            name: 'tutorialhand',
            updateWhenPaused: 0,
            components: [
                sprite,
                behavior
            ]
        });
        var container = new Entity({
            z: settings.z,
            name: settings.name || 'tapHand',
            position: settings.position,
            scale: settings.scale,
            float: settings.float || false,
            rotation: rotation,
            components: [
                entity
            ].concat(settings.components || [])
        }).extend({
            hide: function (onComplete) {
                if (entity.alpha !== 1) {
                    return;
                }

                new Tween({
                    ignoreGameSpeed: true,
                    from: 1,
                    to: 0,
                    in: 20,
                    ease: Tween.LINEAR,
                    onUpdate: function (v, t) {
                        entity.alpha = v;
                    },
                    onComplete: onComplete
                });
            },
            forceStart: function () {
                moveUp();
            }
        });
        return container;
    };
});