/**
 * Level spawner
 * @moduleName LevelSpawner
 * @snippet LevelSpawner|constructor
LevelSpawner({
    assetName: '',
    onComplete: function () {}
})
 */
bento.define('managers/levelspawner', [
    'bento',
    'bento/utils',
    'bento/tiled',
    'bento/math/vector2',
    'bento/math/rectangle',
    'globals',
    'bento/entity',
    'bento/eventsystem',
    'onigiri/onigiri',
    'onigiri/entity3d',
    'entities/polywall',
    'entities/wall',
    'entities/baskethoop',
    'entities/player',
    'entities/coin',
    'entities/enemy',
    'components/manualupdate',
    'entities/drawplane',
    'entities/drawcontroller',
    'entities/key',
    'managers/savevalues',
    'modules/randomizer',
    'entities/platform'
], function (
    Bento,
    Utils,
    Tiled,
    Vector2,
    Rectangle,
    Globals,
    Entity,
    EventSystem,
    Onigiri,
    Entity3D,
    PolyWall,
    Wall,
    BasketHoop,
    Player,
    Coin,
    Enemy,
    ManualUpdate,
    DrawPlane,
    DrawController,
    Key,
    SaveValues,
    Randomizer,
    Platform
) {
    'use strict';
    var LevelSpawner = function (settings) {
        var onComplete = settings.onComplete;
        var assetName = settings.assetName;
        var json = Bento.assets.getJson(assetName);
        var levelScale = parseFloat(Utils.getSafe(json, ['properties', 'levelScale'], 8)) / 6;
        var offset = new Vector2(0, 0);
        var pixelSize = new Rectangle(0, 0, 1, 1);
        var depth = settings.depth || 2;
        var chunk;
        var toOnigiri = function (vec2) {
            // convert bento/box2d 2d vector to three 3d vector
            return new THREE.Vector3(vec2.x * Onigiri.pxToThree, vec2.y * Onigiri.pxToThree, 0);
        };
        var spawn = function () {
            chunk = new Tiled({
                assetName: assetName,
                offset: offset,
                spawnBackground: false,
                attachBackgrounds: false,
                spawnEntities: true,
                attachEntities: true,
                maxCanvasSize: new Vector2(512, 512),
                onInit: function () {
                    Globals.paths = {};
                },
                onObject: function (object, tileSet, tileIndex) {
                    if (!object.gid) {
                        var properties = {};
                        var positionOffset = new Vector2(parseInt(object.x) + offset.x, -parseInt(object.y) - offset.y);
                        if (object.properties) {
                            Utils.forEach(object.properties, function (prop, name) {
                                if (prop.name && prop.value) {
                                    properties[prop.name] = prop.value;
                                } else {
                                    properties[name] = prop;
                                }
                            });
                        }

                        if (object.type === 'Solid') {
                            // spawn walls (without level scale)
                            if (object.polygon) {
                                // add polyWall
                                Bento.objects.attach(new PolyWall({
                                    levelScale: levelScale,
                                    position: positionOffset,
                                    path: object.polygon,
                                    properties: properties
                                }));

                                addPolygon(
                                    positionOffset,
                                    object.polygon
                                );
                            } else {
                                // add normal wall
                                Bento.objects.attach(new Wall({
                                    levelScale: levelScale,
                                    position: positionOffset,
                                    dimension: new Rectangle(0, 0, parseInt(object.width), parseInt(object.height))
                                }));

                                addPolygon(
                                    positionOffset,
                                    rectangleToPolygon(object.width, object.height)
                                );
                            }
                        } else if (object.type === 'Platform') {
                            properties.name = object.name;
                            Bento.objects.attach(new Platform({
                                levelScale: levelScale,
                                position: positionOffset,
                                polygon: object.polygon,
                                dimension: new Rectangle(0, 0, parseInt(object.width), parseInt(object.height)),
                                properties: properties
                            }));
                        } else if (object.type === 'Path') {
                            if (!object.name) {
                                return console.log('Warning: Path has no name!');
                            }
                            if (!object.polyline && !object.polygon) {
                                console.log('Warning: Path has no polyline!');
                                return console.log(object);
                            }
                            var points;
                            if (object.polyline) {
                                points = Utils.cloneJson(object.polyline);
                            } else {
                                points = Utils.cloneJson(object.polygon);
                                if (Utils.isArray(points)) {
                                    points.push(Utils.cloneJson(points[0]));
                                }
                            }
                            Globals.paths[object.name] = {
                                name: object.name,
                                points: points,
                                pixelsPerSecond: properties.pixelsPerSecond,
                                timePerPoint: properties.timePerPoint,
                                ease: properties.ease,
                                levelScale: levelScale,
                            };
                        } else if (properties.module) {
                            bento.require([properties.module], function (Module) {
                                Bento.objects.attach(new Module({
                                    levelScale: levelScale,
                                    position: positionOffset,
                                    dimension: new Rectangle(0, 0, parseInt(object.width), parseInt(object.height)),
                                    tiled: object
                                }));
                            });
                        } else if (object.type === 'Coin') {
                            // spawn other (apply level scale now)
                            positionOffset.x *= levelScale;
                            positionOffset.y *= levelScale;
                            Bento.objects.attach(new Coin({
                                position: toOnigiri(positionOffset),
                                properties: properties
                            }));
                        } else if (object.name.toLowerCase() === 'key') {
                            // spawn other (apply level scale now)
                            positionOffset.x *= levelScale;
                            positionOffset.y *= levelScale;

                            // spawn every 3 levels or so (pre-determined)
                            var randomizer = new Randomizer({
                                seed: SaveValues.currentLevel
                            });
                            if (
                                Utils.getHashParameter('level', '') ||
                                (
                                    randomizer.getRandom(3) === 0 &&
                                    SaveValues.currentLevel > 10
                                ) || (
                                    SaveValues.currentLevel === 4 ||
                                    SaveValues.currentLevel === 6 ||
                                    SaveValues.currentLevel === 8
                                )
                            ) {
                                Bento.objects.attach(new Key({
                                    position: toOnigiri(positionOffset),
                                    properties: properties
                                }));
                            }
                        } else {
                            // spawn other (apply level scale now)
                            positionOffset.x *= levelScale;
                            positionOffset.y *= levelScale;

                            if (object.ellipse && object.name.startsWith('Ring')) {
                                var align = 'center';
                                // the size of the object in Tiled is irrelevant, that's why we use "align" to align the hoop
                                if (object.name.endsWith('right')) {
                                    align = 'right';
                                    positionOffset.x += object.width * levelScale;
                                } else if (object.name.endsWith('left')) {
                                    align = 'left';
                                } else {
                                    positionOffset.x += object.width / 2 * levelScale;
                                }
                                positionOffset.y -= object.height / 2 * levelScale;

                                Bento.objects.attach(new BasketHoop({
                                    ringScale: parseFloat(properties.ringScale || 1),
                                    align: align,
                                    position: toOnigiri(positionOffset)
                                }));
                            }
                            if (object.name.startsWith('Player')) {
                                Bento.objects.attach(new Player({
                                    noBall: properties.noBall || (object.name.toLowerCase().indexOf('-noball') > 0),
                                    direction: parseInt(properties.direction || 1),
                                    position: toOnigiri(positionOffset)
                                }));
                            }
                            if (object.name === 'Enemy') {
                                Bento.objects.attach(new Enemy({
                                    direction: parseInt(properties.direction || 1),
                                    position: toOnigiri(positionOffset)
                                }));
                            }
                        }

                    }
                },
                // onTile: function (tileX, tileY, tilesetJson, tileIndex) {},
                onComplete: function () {
                    Utils.forEach(this.backgrounds, function (instance) {
                        if (instance.name.toLowerCase().indexOf('foreground') >= 0) {
                            instance.z = Globals.Layers.FOREGROUND;
                        }
                        instance.visible = false;
                    });

                    // merge polygons
                    var megaPolygon = polygons[0];
                    Utils.forEach(polygons, function (polygon, i) {
                        if (i !== 0) {
                            megaPolygon = window.PolyBool.union(megaPolygon, polygon);
                        }
                    });
                    // console.log(megaPolygon);
                    Utils.forEach(megaPolygon.regions, function (region) {
                        setupWall(region);
                    });

                    // chekc if there are any tutorials
                    var tutorial = Utils.getSafe(json, ['properties', 'tutorial'], null);
                    if (tutorial && Bento.saveState.load('cheat-skipTutorials', false) === false) {
                        bento.require([tutorial], function (Tutorial) {
                            Bento.objects.attach(new Tutorial({}));
                        });
                    }
                },
                // onLayerMergeCheck: function (layer) {},
                // onLayer: function (layer, index) {},
                onSpawn: function (instance, object, extraData) {
                    // console.log(instance);
                },
                onSpawnComplete: function () {
                    if (onComplete) {
                        onComplete();
                    }
                    EventSystem.fire('levelSpawnComplete');
                }
            });
        };
        /**
         * Set up for the wall visuals: merge all polygons into one and create a geometry
         */
        var polygons = [];
        var rectangleToPolygon = function (width, height) {
            return [{
                x: 0,
                y: 0
            }, {
                x: width,
                y: 0
            }, {
                x: width,
                y: height
            }, {
                x: 0,
                y: height
            }];
        };
        var addPolygon = function (position, shape) {
            // turn into list of coordinates
            var region = [];
            var polygon = {
                regions: [region],
                inverted: false
            };
            Utils.forEach(shape, function (point) {
                region.push([
                    Math.floor(position.x + point.x),
                    Math.floor(-position.y + point.y)
                ]);
            });

            polygons.push(polygon);
        };
        var setupWall = function (path) {
            var wallEntity = new Entity3D({
                name: 'megaWall'
            });
            var getShape = function () {
                var shape = new THREE.Shape();
                var l = path.length;
                var i = 0,
                    x, y;
                var min = new Vector2(Infinity, Infinity);
                var max = new Vector2(-Infinity, -Infinity);

                if (l > 1) {
                    x = path[i][0] * Onigiri.pxToThree * levelScale;
                    // flip y
                    y = -path[i][1] * Onigiri.pxToThree * levelScale;

                    max.x = Math.max(max.x, x);
                    max.y = Math.max(max.y, y);
                    min.x = Math.min(min.x, x);
                    min.y = Math.min(min.y, y);
                    shape.moveTo(x, y);
                    i += 1;
                    while (i < l) {
                        x = path[i][0] * Onigiri.pxToThree * levelScale;
                        // flip y
                        y = -path[i][1] * Onigiri.pxToThree * levelScale;
                        max.x = Math.max(max.x, x);
                        max.y = Math.max(max.y, y);
                        min.x = Math.min(min.x, x);
                        min.y = Math.min(min.y, y);
                        shape.lineTo(x, y);
                        i++;
                    }
                }
                return shape;
            };
            var shape = getShape();

            var extrudeSettings = {
                steps: 8,
                depth: depth * levelScale,
                bevelEnabled: true,
                bevelThickness: 8 * Onigiri.pxToThree,
                bevelSize: 6 * Onigiri.pxToThree,
                bevelOffset: -4 * Onigiri.pxToThree,
                bevelSegments: 7
            };
            // var outlineExtrudeSettings = {...extrudeSettings, ...{bevelSize: 8 * Onigiri.pxToThree}};
            var outlineExtrudeSettings = Utils.cloneJson(extrudeSettings);
            outlineExtrudeSettings.bevelSize = 8 * Onigiri.pxToThree;
            var geometry = new THREE.ExtrudeGeometry(shape, extrudeSettings);
            var outlineGeometry = new THREE.ExtrudeGeometry(shape, outlineExtrudeSettings);
            var outlineMaterial = new THREE.MeshBasicMaterial({
                color: 0x010b13,
                side: THREE.BackSide,
                opacity: 0.5,
                transparent: true
            });
            var material = new THREE.MeshPhongMaterial({
                // color: 0xffffff,
                // emissive: 0x707070,
                // emissiveIntensity: 1,
                // side: THREE.DoubleSide,

                color: Globals.getWallColor(),
                // specular: 0xfefeb6,
                specular: 0xffffff,
                shininess: 4,
                reflectivity: 0.25,

            });
            var mesh = new THREE.Mesh(geometry, material);
            var outlineMesh = new THREE.Mesh(outlineGeometry, outlineMaterial);


            wallEntity.attach(new Entity3D({
                name: 'extrudedMesh',
                position: new THREE.Vector3(0, 0, -depth * levelScale / 2),
                castShadow: true,
                receiveShadow: true,
                object3D: mesh,
                components: [
                    new ManualUpdate({})
                ]
            }));
            wallEntity.attach(new Entity3D({
                name: 'extrudedMeshOutline',
                position: new THREE.Vector3(0, 0, -depth * levelScale / 2),
                castShadow: false,
                receiveShadow: false,
                object3D: outlineMesh,
                components: [
                    new ManualUpdate({})
                ]
            }));


            Bento.objects.attach(wallEntity);
        };

        Globals.currentPuzzle = assetName;

        pixelSize.width = json.tilewidth * json.width;
        pixelSize.height = json.tileheight * json.height;

        // calculate offset to center level (excludes level scale)
        // x = 0 is the middle of the level
        offset.x = -pixelSize.width / 2;
        // push entire level up to put it on the bottom
        offset.y = -pixelSize.height;

        spawn();

        chunk.cameraYOffset = Utils.getSafe(json, ['properties', 'cameraYOffset'], 0);
        chunk.json = json;
        chunk.pixelSize = pixelSize;
        chunk.levelScale = levelScale;
        chunk.depth = depth * levelScale;
        // effective size including levelScale
        chunk.size = new Rectangle(0, 0, pixelSize.width * levelScale * Onigiri.pxToThree, pixelSize.height * levelScale * Onigiri.pxToThree);

        return chunk;
    };
    return LevelSpawner;
});