/**
 * Time manager, manages the time for energy regen and other timed events
 *
 * @moduleName TimeManager
 * @snippet TimeManager.getTime
TimeManager.getTime();
 * @snippet TimeManager.getLocalTime
TimeManager.getLocalTime();
 */
bento.define('managers/timemanager', [
    'bento',
    'utils',
    'bento/entity',
    'bento/eventsystem'
], function (
    Bento,
    Utils,
    Entity,
    EventSystem
) {
    'use strict';

    var started = false;

    var counter = 0;
    var counterTotal = 0;

    var counterBehavior = {
        name: 'counterBehavior',
        start: function () {
            console.log('globalTimer started');
        },
        destroy: function () {
            console.log('Warning: attempted to remove global timer');
            // put self back??
        },
        now: Date.now(),
        update: function (data) {
            var deltaT;
            var now;

            counter++;

            if (counter === 60) {
                // get delta time
                now = Date.now();
                deltaT = now - counterBehavior.now;

                // limit deltaT to 5 seconds to be sure
                if (deltaT > 5000) {
                    deltaT = 5000;
                    //perform a resume event after a DT of > 5 seconds
                    resumeEvent();
                }

                // update now
                counterBehavior.now = now;
                manager.time += deltaT;
                counter = 0;
                counterTotal += 1;
                // save time every 2 seconds
                if (counterTotal % 2 === 0) {
                    manager.saveTime();
                }
            }
        }
    };
    var timerEntity = new Entity({
        z: 0,
        name: 'globalTimer',
        global: true,
        updateWhenPaused: Infinity,
        components: [counterBehavior]
    });

    var initialize = function () {
        if (!started) {
            manager.time = Bento.saveState.load('lastKnownTime', 0);
            manager.previousTime = Bento.saveState.load('lastKnownTime', 0);
            Bento.objects.attach(timerEntity);
            getServerTime(serverComplete, serverFail);


            if (Utils.isDev()) {
                window.timeManager = manager;
            }

            started = true;
        }
    };
    var isRequestingTime = false;
    var getServerTime = function (onComplete, onFail) {
        var dateObject;
        var date;
        var todayNow = Date.now();
        var today = Math.floor(todayNow / 1000 / 60 / 60 / 24);
        var attempt = 0;
        var getAwsTime = function () {
            var xhr = new window.XMLHttpRequest();
            xhr.open("GET", "https://fcqd2ecsh5.execute-api.us-east-1.amazonaws.com/prod/getDate");
            xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4) {
                    if (xhr.status >= 200 && xhr.status < 300) {
                        try {
                            // get date from server
                            dateObject = JSON.parse(xhr.responseText);
                            date = dateObject.date;
                            // calculate new today
                            todayNow = date;
                            today = Math.floor(todayNow / 1000 / 60 / 60 / 24);
                            console.log("Received date from server: " + date);
                            onComplete(date);
                        } catch (e) {
                            if (onFail) {
                                Utils.log('ERROR: could not retrieve or parse date from server.');
                                onFail();
                            }
                        }
                    } else {
                        // should be 200 on success
                        if (onFail) {
                            onFail();
                        }
                    }
                }
            };
            xhr.send();
        };
        var getHeaderTime = function () {
            var dateString;
            var xhr = new window.XMLHttpRequest();
            xhr.open('HEAD', 'https://s3.amazonaws.com/lucky-kat/date.json', true);
            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4) {
                    if (xhr.status >= 200 && xhr.status < 300) {
                        try {
                            // get date from server
                            dateString = this.getResponseHeader("Date");
                            if (dateString.indexOf('GMT') === -1) {
                                dateString += ' GMT';
                            }
                            date = Date.parse(dateString);
                            // calculate new today
                            todayNow = date;
                            today = Math.floor(todayNow / 1000 / 60 / 60 / 24);
                            console.log('Date from server (header): ' + dateString + ' -> ' + date);
                            // headers received
                            onComplete(date);
                        } catch (e) {
                            if (onFail) {
                                Utils.log('ERROR: could not retrieve or parse date from server.' + xhr.responseText);
                                console.log('Backing up with AWS time');
                                getAwsTime();
                            }
                        }
                    } else {
                        if (onFail) {
                            Utils.log('ERROR: bad server response. Backing up with AWS time');
                            getAwsTime();
                        }
                    }
                }
            };
            xhr.send();
        };

        manager.hasUpdatedTime = false;
        if (!isRequestingTime) {
            isRequestingTime = true;
            console.log("Getting server time...");
            if (Utils.WEB || Utils.iDreamSky) {
                attempt = 1;
                getHeaderTime();
            } else {
                if (Utils.isNativeIos() || Utils.isNativeAndroid()) {
                    getHeaderTime();
                } else {
                    getHeaderTime();
                }
            }
        }
    };

    var serverComplete = function (time) {
        isRequestingTime = false;
        manager.online = true;
        manager.time = time;
        manager.hasUpdatedTime = true;
        manager.saveTime();

        // we save the time difference between server and local once
        var timeOffset = Bento.saveState.load('timeOffset', 0);
        if (timeOffset === null) {
            timeOffset = manager.time - Date.now();
            Bento.saveState.save('timeOffset', timeOffset);
        }

        // save the initial playing time (note: not accurate for softlaunch players)
        var initialStartingTime = Bento.saveState.load('initialStartingTime', 0);
        if (initialStartingTime === 0) {
            initialStartingTime = time;
            Bento.saveState.save('initialStartingTime', initialStartingTime);
        }

        // update bootup time
        if (manager.previousTime === 0) {
            manager.previousTime = time;
        }

        EventSystem.fire('serverTimeUpdated');
    };

    var serverFail = function () {
        manager.online = false;
        isRequestingTime = false;

        // try again in 10 seconds
        window.setTimeout(function () {
            if (manager.tryToConnect) {
                manager.updateServerTime();
            }
        }, 10000);
    };

    var getTime = function () {
        return manager.time;
    };

    var resumeEvent = function () {
        manager.hasWarned = false;
        // update now because there can be a big gap between the counterBehavior.now and the real time
        counterBehavior.now = Date.now();
        // update server time
        getServerTime(serverComplete, serverFail);
    };

    var manager = {
        hasWarned: false,
        online: false,
        previousTime: 0,
        time: 0,
        elapsed: 0,
        initialize: initialize,
        getTime: getTime,
        hasUpdatedTime: false,
        tryToConnect: true,
        getLocalTime: function () {
            // retrieves server time corrected with offset
            var offset = Bento.saveState.load('timeOffset', 0) || 0;
            return manager.time - offset;
        },
        resumeEvent: resumeEvent,
        saveTime: function () {
            // save last known time
            Bento.saveState.save('lastKnownTime', manager.time);
        },
        updateServerTime: function (onComplete) {
            // update server time
            getServerTime(function (date) {
                if (onComplete) {
                    onComplete(true);
                }
                serverComplete(date);
            }, function () {
                if (onComplete) {
                    onComplete(false);
                }
                serverFail();
            });
        }
    };
    console.log('return TimeManager')
    return manager;
});