/*
 * The UserStats module tracks commonly used stats for analytics
 * All stats will be saved into Bento.saveState with the "UserStats-" prefix
 *
 * What you need to do:
 * -Init the module (use the snippet)
 * -Fire the event 'UserStats-levelStart' whenever the user starts a level
 * -Fire the event 'UserStats-levelFail' whenever the user fails a level
 * -Fire the event 'UserStats-levelComplete' whenever the user completes a level / levels up
 * -Besides the common stats tracked, you may track custom stats. Do this by sending the event 'UserStats-add'
 *
 * Common Stats:
 * levelUp - a common progression stat, the number of times user leveled up (0-based index)
 *
 * Automatically tracked:
 * initialBootupDate - first time bootup date number
 * initialBootupDay - first time bootup day number
 * lastBootupDay - last day the user booted up 
 * bootups - number of times game booted up
 * sessions - number of sessions the user started
 * totalAdsWatched - total ads watched (interstitial + rewarded)
 * rewardedAdsWatched - amount of rewarded ads watched
 * rewardedAdsWatched-<TAGNAME> - amount of rewarded ads watched per tag
 * interstitialAdsWatched - amount of interstitial ads watched
 * retention - internally tracked retention as JSON object (note: uses Date.now, which may be affected by system settings)
 * dayActive - array of days that user was active
 * totalPlayTime - total time played (in ms)
 * 
 * @snippet EventSystem.fire()|UserStats-levelStart
EventSystem.fire('UserStats-levelStart', {
    level: ${1:levelNumber}
});
 * @snippet EventSystem.fire()|UserStats-levelComplete
EventSystem.fire('UserStats-levelComplete', {
    level: ${1:levelNumber}
});
 * @snippet EventSystem.fire()|UserStats-levelFail
EventSystem.fire('UserStats-levelFail', {
    level: ${1:levelNumber}
});
 * @snippet EventSystem.fire()|UserStats-new
EventSystem.fire('UserStats-add', {
    name: '${1:nameOfStat}',
    value: 1
});
 * @moduleName UserStats
 */
bento.define('managers/userstats', [
    'bento',
    'bento/eventsystem',
    'bento/utils'
], function (
    Bento,
    EventSystem,
    Utils
) {
    'use strict';
    var UserStats = new Object({
        sessionStartTime: Date.now(),
        /**
         * @snippet UserStats.init()|snippet
UserStats.init({
    onFirstBootup: function () {},
    onBootup: function (data) {
        // send retention/progress events
        var retainedDay = data.retainedDay;
        var day = data.day;
        var lastBootupDay = data.lastBootupDay;
        var level = Bento.saveState.load('UserStats-levelUp', 0);
    
        // send only if it's a new day
        if (lastBootupDay !== day) {
            // send to Tenjin
            if (retainedDay === 1 || retainedDay === 7 || retainedDay === 14  || retainedDay === 30) {
                EventSystem.fire('Tenjin', {
                    name: 'D' + retainedDay + '_retention'
                });
            }
            // send to GA
            if ((retainedDay >= 1 && retainedDay <= 7) || retainedDay === 14  || retainedDay === 30) {
                EventSystem.fire('GameAnalytics-addDesignEvent', {
                    eventId: 'LevelPerRetention:D' + retainedDay,
                    value: (level + 1)
                });
            }
        }
    },
    onInterstitialAdComplete: function (data) {
        data = data || {};
        // send extra events to Tenjin
        if (data.count === 10) {
            EventSystem.fire('Tenjin', {
                name: '10_is_watched'
            });
        }
        // send extra events to Tenjin
        if (data.count === 20) {
            EventSystem.fire('Tenjin', {
                name: '20_is_watched'
            });
        }
        // send extra events to Tenjin
        if (data.count === 50) {
            EventSystem.fire('Tenjin', {
                name: '50_is_watched'
            });
        }
        // send extra events to Tenjin
        if (data.count === 100) {
            EventSystem.fire('Tenjin', {
                name: '100_is_watched'
            });
        }
    },
    onRewardedAdComplete: function (data) {
        data = data || {};
        // send extra events to Tenjin
        if (data.count === 10) {
            EventSystem.fire('Tenjin', {
                name: '10_rv_watched'
            });
        }
        if (data.count === 20) {
            EventSystem.fire('Tenjin', {
                name: '20_rv_watched'
            });
        }
        if (data.count === 50) {
            EventSystem.fire('Tenjin', {
                name: '50_rv_watched'
            });
        }
        if (data.count === 100) {
            EventSystem.fire('Tenjin', {
                name: '100_rv_watched'
            });
        }
    },
    onLevelStart: function (data) {
        // trigger this with EventSystem.fire('UserStats-levelStart', {level: levelNumber});
        // send extra ads/level events to GameAnalytics 
        var level = data.level || 0;
    
        // GA progression event
        EventSystem.fire('GameAnalytics-addProgressionEvent', {
            progressionStatus: 1, // Start
            progression01: "level_" + level
        });
    },
    onLevelFail: function (data) {
        // trigger this with EventSystem.fire('UserStats-levelFail', {level: levelNumber});
        var level = data.level || 0;

        // GA progression event
        EventSystem.fire('GameAnalytics-addProgressionEvent', {
            progressionStatus: 3, // Fail
            progression01: "level_" + level
        });
    },
    onLevelComplete: function (data) {
        // trigger this with EventSystem.fire('UserStats-levelComplete', {level: levelNumber});
        // send extra ads/level events to GameAnalytics 
        var level = data.level || 0;
    
        var rewardedAdsWatched = Bento.saveState.load('UserStats-rewardedAdsWatched', 0);
        
        // GA design event: number of ads watched per level
        EventSystem.fire('GameAnalytics-addDesignEvent', {
            eventId: 'AdsPerLevel:Level:' + level,
            value: rewardedAdsWatched
        });
    
        // GA progression event
        EventSystem.fire('GameAnalytics-addProgressionEvent', {
            progressionStatus: GameAnalytics.EGAProgressionStatus.Complete,
            progression01: "level_" + level
        });
    }
})
         */
        init: function (settings) {
            var onBootup = settings.onBootup;
            var onFirstBootup = settings.onFirstBootup;
            var onRewardedAdComplete = settings.onRewardedAdComplete;
            var onInterstitialAdComplete = settings.onInterstitialAdComplete;
            var onLevelComplete = settings.onLevelComplete;
            var onLevelStart = settings.onLevelStart;
            var onLevelFail = settings.onLevelFail;
            var now = Date.now();
            var day = Math.floor(now / 1000 / 60 / 60 / 24);
            var dayDiff = 0;
            var retainedDays = Bento.saveState.load('UserStats-retainedDays', {});
            var lastBootupDay = Bento.saveState.load('UserStats-lastBootupDay', 0);
            var onLevelUp = function (data) {
                var level = data.level;
                Bento.saveState.save('UserStats-levelUp', level);
                if (onLevelComplete) {
                    onLevelComplete({
                        level: level
                    });
                }
            };

            /** 
             * Bootup analytics
             */
            if (Bento.saveState.load('UserStats-bootups', 0) === 0) {
                Bento.saveState.save('UserStats-initialBootupDate', now);
                Bento.saveState.save('UserStats-initialBootupDay', day);

                if (onFirstBootup) {
                    onFirstBootup();
                }
            }
            // assuming init is only called once in a lifetime, add to the bootups/sessions stat
            // note: difference between bootups and session: session is also counted when user puts the game in suspend mode
            // whereas bootup is only once when the game is booted up 
            Bento.saveState.add('UserStats-bootups', 1);
            Bento.saveState.add('UserStats-sessions', 1);

            // retention (based on date change)
            dayDiff = day - Bento.saveState.load('UserStats-initialBootupDay', day);
            UserStats.retainedDay = dayDiff;
            retainedDays[dayDiff] = 1;
            Bento.saveState.save('UserStats-retainedDays', retainedDays);

            // callback
            if (onBootup) {
                onBootup({
                    retainedDay: dayDiff,
                    day: day,
                    lastBootupDay: lastBootupDay
                });
            }

            /** 
             * Count sessions
             */
            if (document && document.addEventListener) {
                document.addEventListener('resume', UserStats.onResume, false);
            }

            /**
             * Count active days
             */
            if (lastBootupDay !== day) {
                var activeDays = Bento.saveState.load('UserStats-activeDays', []);
                activeDays.push(day);
                Bento.saveState.save('UserStats-activeDays', activeDays);
            }

            /**
             * Count ads watched
             */
            EventSystem.on('onRewardedAdComplete', function (data) {
                Bento.saveState.add('UserStats-totalAdsWatched', 1);
                // total rewarded
                Bento.saveState.add('UserStats-rewardedAdsWatched', 1);

                // per tag
                if (data.tag) {
                    Bento.saveState.add('UserStats-rewardedAdsTag-' + data.tag, 1);
                }

                if (onRewardedAdComplete) {
                    onRewardedAdComplete({
                        count: Bento.saveState.load('UserStats-rewardedAdsWatched', 0),
                        tag: data.tag
                    });
                }
            });
            EventSystem.on('onInterstitialAdComplete', function (data) {
                Bento.saveState.add('UserStats-totalAdsWatched', 1);
                // total interstitials
                Bento.saveState.add('UserStats-interstitialAdsWatched', 1);

                // per tag
                if (data.tag) {
                    Bento.saveState.add('UserStats-interstitialAdsTag-' + data.tag, 1);
                }

                if (onInterstitialAdComplete) {
                    onInterstitialAdComplete({
                        count: Bento.saveState.load('UserStats-interstitialAdsWatched', 0),
                        tag: data.tag
                    });
                }

            });

            /**
             * Level up
             */
            EventSystem.on('UserStats-levelUp', onLevelUp); // legacy name
            EventSystem.on('UserStats-levelComplete', onLevelUp);
            EventSystem.on('UserStats-levelStart', onLevelStart);
            EventSystem.on('UserStats-levelFail', onLevelFail);

            /**
             * Custom stats
             */
            EventSystem.on('UserStats-add', UserStats.onAddStat);

            /**
             * Track playtime
             */
            if (!Utils.isNativeIos()) {
                document.addEventListener('pause', function () {
                    var diff = Date.now() - UserStats.sessionStartTime;
                    Bento.saveState.add('UserState-totalPlayTime', diff);

                    // reset session start time
                    UserStats.sessionStartTime = Date.now();
                }, false);
            } else {
                // this event is for iOS only and is more accurate than pause
                document.addEventListener('resign', function () {
                    var diff = Date.now() - UserStats.sessionStartTime;
                    Bento.saveState.add('UserState-totalPlayTime', diff);

                    // reset session start time
                    UserStats.sessionStartTime = Date.now();
                }, false);

            }

            // update lastBootupDay, keep this below
            lastBootupDay = day;
            Bento.saveState.save('UserStats-lastBootupDay', lastBootupDay);
        },
        onAddStat: function (data) {
            Bento.saveState.add('UserStats-' + data.name, data.addedValue);
        },
        onResume: function () {
            Bento.saveState.add('UserStats-sessions', 1);

            // reset session start time
            UserStats.sessionStartTime = Date.now();
        }
    });
    return UserStats;
});