/**
 * AB test manager. Instructions:
 * Use ABTest.addTest() before you init Game Analytics 
 * Use ABTest.init() after you init GameAnalytics
 *
 * Example:
GameAnalytics.configureBuild(window.VERSION);

// Configure Custom Dimension 01
ABTest.setTest(1, 'SomeTest', ['Off', 'On']);
// Configure Custom Dimension 02
ABTest.setTest(2, 'OtherTest', ['Yes', 'No', 'Maybe']);
// Configure Custom Dimension 03
ABTest.setTest(3, 'AnotherOne', ['A', 'B']);

// Force concluded tests
ABTest.concludeTest('OldTest', 'On');

GameAnalytics.initialize({
    gameKey: '123abc',
    secretKey: '321cba'
});

// Set Custom Dimensions
ABTest.init();

// ...

// Check values later in game
if (ABTest.getOption('SomeTest') === 'On') {
    // do stuff...
}

 *
 *
 * 
 * @moduleName ABTest
 */
bento.define('modules/abtest', [
    'bento',
    'bento/eventsystem',
    'bento/utils',
    'modules/gameanalytics'
], function (
    Bento,
    EventSystem,
    Utils,
    GameAnalytics
) {
    'use strict';
    var _keys = [];
    var _options = [];
    /**
     * For sanity, we always call this even if the custom dimension value didn't change since last boot.
     */
    var setCustomDimensions = function () {
        for (var ii = 0; ii < _keys.length; ii++) {
            var idx = ii + 1;
            var option = Bento.saveState.load('abTest-' + _keys[ii], null);
            var parsedOption = _keys[ii] + ':' + option;
            if (option !== null) {
                EventSystem.fire('GameAnalytics-setCustomDimension0' + idx, parsedOption);
            }
        }
    };
    var configureTest = function (customDimension, key, options, parsedOptions) {
        var saved = Bento.saveState.load('abTest-' + key, null);

        // configure the custom dimension
        if (customDimension === 1) {
            GameAnalytics.configureAvailableCustomDimensions01(parsedOptions);
        } else if (customDimension === 2) {
            GameAnalytics.configureAvailableCustomDimensions02(parsedOptions);
        } else if (customDimension === 3) {
            GameAnalytics.configureAvailableCustomDimensions03(parsedOptions);
        } else {
            throw "Cannot set CustomDimensions0" + customDimension;
        }
        // save the test in saveState
        if (saved) {
            // already saved value for this, done
            console.log("A/B test '" + key + "' was loaded as " + saved);
            return;
        }
        var randomized = Utils.pickRandomFrom(options);
        Bento.saveState.save('abTest-' + key, randomized);
        console.log("A/B test '" + key + "' is set to " + randomized);
    };
    // throws if any setting is invalid
    var checkStrings = function (key, options) {
        Utils.forEach(options, function (option) {
            var name = key + ':' + option;
            if (name.length > 32) {
                throw 'CustomDimension name can not be longer than 32 characters  -> "' + name + '" (' + name.length + ' characters)';
            }
            if (option === null || option === void(0)) {
                throw 'ERROR: null or undefined as option is not valid';
            }
            if (option.toString() === '') {
                throw 'ERROR: empty strings are not valid';
            }
        });

    };
    var ABTest = {
        /**
         * @snippet ABtest.init()|postGaInit
        ABtest.init()
         */
        init: function (onComplete) {
            setCustomDimensions();

            if (onComplete) {
                onComplete();
            }
        },
        
        // DEPRECATED
        addTest: function (key, options) {
            if (_keys.length >= 3) {
                throw 'Can set 3 custom dimensions maximum';
            }
            checkStrings(key, options);

            _keys.push(key);
            _options.push(options);

            var parsedOptions = options.map(function (o) {
                return key + ':' + o;
            });

            configureTest(_keys.length, key, options, parsedOptions);
        },
        /**
         * @snippet ABTest.setTest()|preGaInit
        ABTest.setTest(${1:1}, '${2:AB_TEST_KEY}', [${3:'Option1', 'Option2', ...}])
         */
        setTest: function (customDimension, key, options) {
            var index = parseInt(customDimension) - 1;

            checkStrings(key, options);

            _keys[index] = key;
            _options[index] = options;

            var parsedOptions = options.map(function (o) {
                return key + ':' + o;
            });
            configureTest(customDimension, key, options, parsedOptions);
        },
        /**
         * @snippet ABTest.getOption()|String
        ABTest.getOption('${1:AB_TEST_KEY}')
         */
        getOption: function (key) {
            return Bento.saveState.load('abTest-' + key, null);
        },
        /**
         * @snippet ABTest.setOption()|Snippet
        ABTest.setOption('${1:AB_TEST_KEY}', '${2:Option}')
         */
        setOption: function (key, option, ignoreCustomDimension) {
            var idx = _keys.indexOf(key);

            if (!ignoreCustomDimension) {
                // use ignoreCustomDimension after A/B test is concluded
                if (idx === -1) {
                    console.warn('WARNING: test ' + key + ' is not an active test, so it can not be set as Custom Dimension.');
                } else {
                    var parsedOption = key + ':' + option;
                    EventSystem.fire('GameAnalytics-setCustomDimension0' + (idx + 1), parsedOption);
                }
            }
            Bento.saveState.save('abTest-' + key, option);

        },
        /**
         * @snippet ABTest.concludeTest()|Snippet
        ABTest.concludeTest('${1:AB_TEST_KEY}', '${2:Option}')
         */
        concludeTest: function (key, option) {
            // concludeTest is an alias for setOption, but ignores the custom dimension
            ABTest.setOption(key, option, true);
        },

        /**
         * @snippet ABTest.getTests()|ArrayOfObjects
        ABTest.getTests()
         */
        getTests: function () {
            // returns active tests
            var tests = [];
            for (var ii = 0; ii < _keys.length; ii++) {
                tests.push({
                    key: _keys[ii],
                    options: _options[ii],
                    current: Bento.saveState.load('abTest-' + _keys[ii], null),
                });
            }
            return tests;
        }
    };

    return ABTest;
});