/**
 * Box2D wrapper module. Compatible with Box2DWeb and
 * @moduleName Box2D
 *
 * Usage:
 * Add to your game's index.html:

    <script type="text/javascript" src="lib/box2dweb.js"></script>
    <script type="text/javascript" src="lib/cocoonbox2d.js"></script>

 *
 * Most of the time, you want to use the wrapper functions provided by this module.
 * If you want to go under the hood, see the Box2DFlash docs:
 *   - http://www.box2dflash.org/docs/2.1a/reference/
 *
 * But beware, Cocoon Box2D is missing A LOT of features.
 * Check the source to see if what you want is implemented:
 *   - https://bitbucket.org/ludei/cocoon_cocoonjsextensions/src/master/src/box2d_cocoonJS.js
 *
 * Also the LÖVE physics docs are helpful, but beware API differences:
 *   - https://love2d.org/wiki/love.physics
 *
 * Create a body:
 *
 * Handling collisions:
 *
 * Using sensors:
 *
 * Limitations:
 * - There is only 1 physics world. If you call Box2D.setup() again, the world will be replaced.
 * - You must go under the hood to work with joints, or make on-the-fly changes to shapes and fixtures.
 *
 * @snippet Box2D.setup
Box2D.setup({
    gravity: new Vector2(0, Box2D.toPixels(9.8)),
    canSleep: true,
    debugDraw: false,
    debugLayer: 999999999999,
});
Bento.objects.attach(Box2D);

 *
 * @snippet Box2D.Body
Box2D.Body({
    type: 'dynamic',
    fixture: {
        shape: new Box2D.BoxShape(12, 14),
        friction: 1,
        restitution: 0.5,
        density: 0.3,
    }
})
 */
bento.define('modules/box2d', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/packedimage'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    PackedImage
) {
    'use strict';
    var Box2D = window.Box2D;

    var box2DWorld;
    var meterToPixels = 1; // 1 meter = 1
    var pixelsToMeter = 1 / meterToPixels;
    var canSleep = true;
    var clearForces = true;
    var isSimulating = false;
    var deferedFunctions = [];

    // distance / vector conversions
    var toPixels = function (m) {
        return meterToPixels * m;
    };
    var toMeters = function (px) {
        return pixelsToMeter * px;
    };
    var toBox2D = function (vec2) {
        return new Box2D.Common.Math.b2Vec2(pixelsToMeter * vec2.x, pixelsToMeter * vec2.y);
    };
    var toBento = function (vec2) {
        return new Vector2(meterToPixels * vec2.x, meterToPixels * vec2.y);
    };

    // conversion without fixing units
    var toVector2 = function (b2Vec2) {
        return new Vector2(b2Vec2.x, b2Vec2.y);
    };
    var toB2Vec2 = function (vec2) {
        return new Box2D.Common.Math.b2Vec2(vec2.x, vec2.y);
    };

    // enums
    var bodyTypes = {
        'static': Box2D.Dynamics.b2Body.b2_staticBody,
        'dynamic': Box2D.Dynamics.b2Body.b2_dynamicBody,
        'kinematic': Box2D.Dynamics.b2Body.b2_kinematicBody,
    };

    // potationally <- typo of the day

    // potentially a way to optimise family checking
    // since cocoon box2d doesn't support collision filters

    // var familyBits = {
    //     // 'name': 0x0001   etc.
    // };
    // var familyBitsCount = 1;
    // var getFamilyBits = function (family) {
    //     var bits = familyBits[family];
    //     if (!bits) {
    //         bits = familyBitsCount
    //         familyBits[family]
    //     }
    // };

    // debug drawing
    var debugDraw = new Box2D.Dynamics.b2DebugDraw();
    var canvas;
    var context;
    var debugSprite;
    var debugEntity;
    var debugLayer;

    var Box2DModule = {
        z: 0,
        name: 'Box2D',
        pxToM: pixelsToMeter,
        mToPx: meterToPixels,
    
        gravity: new Box2D.Common.Math.b2Vec2(0, 9.8),

        toPixels: toPixels,
        toMeters: toMeters,
        toVector2: toVector2,
        toB2Vec2: toB2Vec2,
        toBox2D: toBox2D,
        toBento: toBento,

        Box2D: Box2D,
        box2DWorld: box2DWorld,
        debugDraw: false,
        debugAlpha: 1,

        /**
         * Is Box2d currently inside a simulation step?
         */
        isSimulating: function () {
            return isSimulating;
        },
        /**
         * Push a function that needs to be ran after the simulation
         */
        defer: function (fn) {
            deferedFunctions.push(fn);
        },
        /**
         * Setup box2d world
         */
        setup: function (settings) {
            if (Utils.isDefined(settings.gravity)) {
                Box2DModule.gravity = toBox2D(settings.gravity);
            }
            if (Utils.isDefined(settings.canSleep)) {
                canSleep = settings.canSleep;
            }
            if (Utils.isDefined(settings.z)) {
                this.z = settings.z;
            }
            if (Utils.isDefined(settings.meterToPixels)) {
                meterToPixels = settings.meterToPixels;
            }
            pixelsToMeter = 1 / meterToPixels;
            this.pxToM = pixelsToMeter;
            this.mToPx = meterToPixels;
            this.debugDraw = settings.debugDraw;
            this.debugAlpha = settings.debugAlpha;
            debugLayer = settings.debugLayer;

            if (settings.addNow) {
                Bento.objects.attach(Box2DModule);
            }
            return Box2DModule;

            // Box2DModule.logAvailableApi();
        },
        start: function (data) {
            // setup debug drawing
            var viewport = Bento.getViewport();
            var b2DebugDraw = Box2D.Dynamics.b2DebugDraw;

            // setup world
            box2DWorld = new Box2D.Dynamics.b2World(Box2DModule.gravity, canSleep);
            box2DWorld.SetContinuousPhysics(true);
            this.box2DWorld = box2DWorld;

            // collision handlers
            // TODO: pass to EventSystem?

            // for any contact event, we check the fixture Settings object for collision handlers
            // if we find a handler, call it with the other fixture's Settings object as the first argument

            var listener = new Box2D.Dynamics.b2ContactListener();

            // CocoonJS only calls BeginContact  :(
            listener.BeginContact = function (contact) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onBeginContact) {
                    a.onBeginContact(b, contact);
                }
                if (b.onBeginContact) {
                    b.onBeginContact(a, contact);
                }
            };

            listener.EndContact = function (contact) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onEndContact) {
                    a.onEndContact(b, contact);
                }
                if (b.onEndContact) {
                    b.onEndContact(a, contact);
                }
            };
            listener.PostSolve = function (contact, impulse) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onPostSolve) {
                    a.onPostSolve(b, contact, impulse);
                }
                if (b.onPostSolve) {
                    b.onPostSolve(a, contact, impulse);
                }
            };
            listener.PreSolve = function (contact, oldManifold) {
                var a = contact.GetFixtureA().GetUserData();
                var b = contact.GetFixtureB().GetUserData();
                if (a.onPreSolve) {
                    a.onPreSolve(b, contact, oldManifold);
                }
                if (b.onPreSolve) {
                    b.onPreSolve(a, contact, oldManifold);
                }
            };
            box2DWorld.SetContactListener(listener);

            // var filter = new Box2D.Dynamics.b2ContactFilter();
            // filter.ShouldCollide = function (fixtureA, fixtureB) {
            //     var a = fixtureA.GetUserData();
            //     var b = fixtureB.GetUserData();
            //     return (a.onFilter ? a.onFilter(b) : true) && (b.onFilter ? b.onFilter(a) : true);
            // };
            // box2DWorld.SetContactFilter(filter);

            // debug drawing
            if (!this.debugDraw) {
                return;
            }
            canvas = document.createElement('canvas');
            canvas.width = viewport.width;
            canvas.height = viewport.height;
            context = canvas.getContext('2d');
            debugSprite = new Sprite({
                image: new PackedImage(canvas)
            });
            debugEntity = new Entity({
                z: debugLayer || 1000000000000,
                name: 'Box2DDebug',
                alpha: this.debugAlpha,
                components: [debugSprite],
                // float: true
            });
            debugDraw.SetSprite(context);
            debugDraw.SetDrawScale(meterToPixels);
            debugDraw.SetFlags(b2DebugDraw.e_shapeBit);
            box2DWorld.SetDebugDraw(debugDraw);
            Bento.objects.attach(debugEntity);
        },
        update: function (data) {
            // step box2d simulation
            isSimulating = true;
            box2DWorld.Step(1 / 60 * data.speed, 10, 10);

            if (clearForces) {
                box2DWorld.ClearForces();
            }

            isSimulating = false;
            // run defered functions
            Utils.forEach(deferedFunctions, function (fn, i, l, breakLoop) {
                fn();
            });
            // clean up
            deferedFunctions = [];
        },
        draw: function (data) {
            if (this.debugDraw) {
                canvas.texture = null;
                box2DWorld.DrawDebugData();
            }
        },
        RevoluteJoint: function (data) {
            var jointDef = new Box2D.Dynamics.Joints.b2RevoluteJointDef();
            jointDef.Initialize(data.bodyA.getB2Body(), data.bodyB.getB2Body(), toBox2D(data.anchor));
            jointDef.lowerAngle = data.lowerAngle;
            jointDef.upperAngle = data.upperAngle;
            jointDef.maxMotorTorque = data.maxMotorTorque;
            jointDef.motorSpeed = data.motorSpeed;
            jointDef.enableLimit = data.enableLimit;
            jointDef.enableMotor = data.enableMotor;
            var joint = box2DWorld.CreateJoint(jointDef);
            return joint;
        },

        // shorthands - do we need wrapper classes for these?
        CircleShape: function (radius, center) {
            var shape = new Box2D.Collision.Shapes.b2CircleShape(toMeters(radius));
            // cocoon doesn't support center
            if (center) {
                if (window.nativeBox2D) {
                    Utils.log('SetLocalPosition not supported.');
                } else {
                    shape.SetLocalPosition(toBox2D(center));
                }
            }
            return shape;
        },
        BoxShape: function (halfW, halfH, center) {
            var shape = new Box2D.Collision.Shapes.b2PolygonShape();
            halfW = toMeters(halfW);
            halfH = toMeters(halfH);
            var ox = center ? toMeters(center.x) : 0;
            var oy = center ? toMeters(center.y) : 0;
            if (center) {
                var points = [{
                    x: -halfW + ox,
                    y: -halfH + oy
                }, {
                    x: halfW + ox,
                    y: -halfH + oy
                }, {
                    x: halfW + ox,
                    y: halfH + oy
                }, {
                    x: -halfW + ox,
                    y: halfH + oy
                }, ];
                shape.SetAsArray(points, points.length);
            } else {
                shape.SetAsBox(halfW, halfH);
            }
            return shape;
        },
        PolygonShape: function (points) {
            var len = points.length;
            var shape = new Box2D.Collision.Shapes.b2PolygonShape();
            for (var i = 0; i < len; i++) {
                var p = points[i];
                p.x = toMeters(p.x);
                p.y = toMeters(p.y);
            }
            shape.SetAsArray(points, len);
            return shape;
        },


        // category:   what type is this fixture?
        // mask:       what types does this fixture collide with?
        // group:      two fixtures must have the same group in order to collide?
        Filter: function (category, mask, group) {
            var filter;
            if (Box2D.Dynamics.b2FilterData) {
                filter = new Box2D.Dynamics.b2FilterData();
            } else {
                // cocoon
                filter = {
                    categoryBits: 1,
                    maskBits: 0xffff,
                    groupIndex: 0
                };
            }
            if (category != null) {
                filter.categoryBits = category;
            }
            if (mask != null) {
                filter.maskBits = mask;
            }
            if (group != null) {
                filter.groupIndex = group;
            }
            return filter;
        },

        // Wrapper component around b2Body -- attach this to an entity to hijack position and rotation
        Body: function (settings) {
            var entity;
            var box2DBody;
            var hasStarted = false;

            var component = {
                name: 'body',

                fixtures: null,

                start: function (data) {
                    var fixtures;
                    var bodyDef = new Box2D.Dynamics.b2BodyDef();
                    var setBodyDef = function (name) {
                        // sets body property if existant in settings
                        if (Utils.isDefined(settings[name])) {
                            bodyDef[name] = settings[name];
                        }
                    };
                    if (!entity.isAdded) {
                        return;
                    }

                    // body settings
                    bodyDef.type = settings.type ? bodyTypes[settings.type] : Box2D.Dynamics.b2Body.b2_staticBody;
                    bodyDef.position.Set(
                        toMeters(entity.position.x),
                        toMeters(entity.position.y)
                    );
                    if (entity.euler) {
                        // bodyDef.angle = entity.euler.z;
                    } else {
                        bodyDef.angle = entity.rotation;
                    }
                    setBodyDef('active');
                    setBodyDef('allowSleep');
                    setBodyDef('angularDamping');
                    setBodyDef('angularVelocity');
                    setBodyDef('awake');
                    setBodyDef('bullet');
                    setBodyDef('fixedRotation');
                    setBodyDef('inertiaScale');
                    setBodyDef('linearDamping');
                    setBodyDef('linearVelocity');

                    box2DBody = box2DWorld.CreateBody(bodyDef);

                    if (box2DBody === null && Box2DModule.isSimulating()) {
                        // TODO: defer start function automatically
                        console.error('Could not instantiate Body. Use Box2D.defer to instantiate a body outside the Box2D simulation step.');
                        return;
                    }

                    // fixture settings
                    fixtures = settings.fixture || settings.fixtures || [];
                    if (!Utils.isArray(fixtures)) {
                        fixtures = [fixtures];
                    }
                    Utils.forEach(fixtures, function (fixture, i, l, breakLoop) {
                        component.createFixture(fixture);
                    });

                    component.fixtures = fixtures;

                    // userdata is reference to component
                    box2DBody.SetUserData(this);
                    hasStarted = true;
                },
                destroy: function (data) {
                    if (Box2DModule.isSimulating()) {
                        // wait a moment until Box2D is done simulating
                        Box2DModule.defer(function () {
                            box2DWorld.DestroyBody(box2DBody);
                            hasStarted = false;
                        });
                    } else {
                        // can safely remove
                        box2DWorld.DestroyBody(box2DBody);
                        hasStarted = false;
                    }
                },
                update: function (data) {
                    // update transform to match rigid body
                    var pos = box2DBody.GetPosition();
                    entity.position.x = toPixels(pos.x);
                    entity.position.y = toPixels(pos.y);
                    if (!settings.ignoreRotation) {
                        entity.rotation = box2DBody.GetAngle();
                        if (entity.euler) {
                            entity.euler.z = box2DBody.GetAngle();
                        }
                    }
                },
                draw: function (data) {},
                attached: function (data) {
                    entity = data.entity;
                },
                getB2Body: function () {
                    if (hasStarted === false) {
                        console.error('Body is not initialized yet. Wait until entity start');
                        return null;
                    }
                    return box2DBody;
                },


                // wrapper methods which add / remove fixtures using a 'settings' object
                createFixture: function (fixture) {
                    var fixtureDef = new Box2D.Dynamics.b2FixtureDef();
                    var setFixtureDef = function (name) {
                        // sets fixture property if existant in settings
                        if (Utils.isDefined(fixture[name])) {
                            fixtureDef[name] = fixture[name];
                        }
                    };
                    fixtureDef.shape = fixture.shape || new Box2D.Collision.Shapes.b2CircleShape(toMeters(16));
                    setFixtureDef('friction');
                    setFixtureDef('restitution');
                    setFixtureDef('density');
                    setFixtureDef('isSensor');
                    setFixtureDef('filter');

                    // fixture settings act as user data, for use in collision handlers.
                    // body and entity fields are added to the settings for convenience.
                    // and we need to keep a reference to the underlying fixture in case we want to remove it later.
                    fixtureDef.userData = fixture;
                    fixture.body = component;
                    fixture.entity = entity;
                    fixture.box2DFixture = box2DBody.CreateFixture(fixtureDef);
                    if (component.fixtures) {
                        component.fixtures.push(fixture);
                    }
                },

                destroyFixture: function (fixture) {
                    var index = component.fixtures.indexOf(fixture);
                    if (index === -1) {
                        return;
                    }
                    box2DBody.DestroyFixture(fixture.box2DFixture);
                    component.fixtures.splice(index, 1);
                },

                // wrapper methods that automatically convert between Vector2 / b2Vec2

                // NOTE: there's so many of these...
                // it makes me wonder about what the performance overhead is like when creating closures?
                // My understanding is that prototypes would lead to smaller creation overhead and less memory usage?

                setPositionAndAngle: function (pos, angle) {
                    box2DBody.setPositionAndAngle(toBox2D(pos), angle);
                },
                getAngle: function () {
                    return box2DBody.GetAngle();
                },
                setAngle: function (angle) {
                    box2DBody.SetAngle(angle);
                },
                getPosition: function () {
                    return toBento(box2DBody.GetPosition());
                },
                setPosition: function (pos) {
                    box2DBody.SetPosition(toBox2D(pos));
                },
                getWorldCenter: function () {
                    return toBento(box2DBody.GetWorldCenter());
                },
                getLocalCenter: function () {
                    return toBento(box2DBody.GetLocalCenter());
                },
                getLinearVelocity: function () {
                    return toBento(box2DBody.GetLinearVelocity());
                },
                setLinearVelocity: function (vec) {
                    box2DBody.SetLinearVelocity(toBox2D(vec));
                },
                getAngularVelocity: function () {
                    return box2DBody.GetAngularVelocity();
                },
                setAngularVelocity: function (omega) {
                    box2DBody.SetAngularVelocity(omega);
                },
                applyForce: function (force, point) {
                    box2DBody.ApplyForce(toBox2D(force), toBox2D(point));
                },

                // this function is missing from box2dweb so we implement it ourselves
                // applyAngularImpulse: function (impulse) {

                //     // impulse is in kg*m^2/s, so we need to scale down twice
                //     impulse = toMeters(toMeters(impulse));

                //     if (box2DBody.m_type !== Box2D.Dynamics.b2Body.b2_dynamicBody) {
                //         return;
                //     }
                //     if (!box2DBody.IsAwake()) {
                //         box2DBody.SetAwake(true);
                //     }
                //     box2DBody.m_angularVelocity += box2DBody.m_invI*impulse;
                // },
                applyTorque: function (torque) {
                    box2DBody.ApplyTorque(toMeters(toMeters(torque)));
                },
                applyImpulse: function (impulse, point) {
                    box2DBody.ApplyImpulse(toBox2D(impulse), toBox2D(point));
                },
                // split(), merge()  // too low level for this wrapper?
                getMass: function () {
                    return box2DBody.GetMass();
                },
                // getInertia: function () {
                //     var I = 0.0;
                //     component.fixtures.forEach(function (f) {
                //         var density = f.GetDensity();
                //         var shape = f.GetShape();
                //         if (density === 0.0) {
                //             return; // continue
                //         }
                //         switch (shape.GetType()) {
                //             case b2Shape.e_circleShape:
                //                 var radius = shape.GetRadius();
                //                 var mass = density * Math.PI * radius * radius;
                //                 // I += mass * (0.5 * radius * radius + )
                //                 break;
                //             case b2Shape.e_polygonShape:
                //                 break;
                //             default:
                //                 Utils.log('ERROR: unknown fixture type: '+shape.GetType());
                //         }
                //     });
                //     return toPixels(toPixels(box2DBody.GetInertia()));
                // },
                // getMassData(), setMassData(), resetMassData()
                getWorldPoint: function (localPoint) {
                    return toBento(box2DBody.GetWorldPoint(toBox2D(localPoint)));
                },
                getWorldVector: function (localVector) {
                    // doesn't exist in cocoon so we implement it ourselves

                    // sanity check
                    if (Utils.isDev() && !Utils.isCocoonJs()) {
                        var a = toBento(box2DBody.GetWorldVector(toBox2D(localVector)));
                        var b = localVector.clone().rotateRadian(box2DBody.GetAngle());
                        if (Math.abs(a.x - b.x) > 0.00000001 || Math.abs(a.y - b.y) > 0.00000001) {
                            console.log('getWorldVector wrong result! expected:', a, 'actual:', b);
                        }
                    }

                    return localVector.clone().rotateRadian(box2DBody.GetAngle());
                },
                getLocalPoint: function (worldPoint) {
                    return toBento(box2DBody.GetLocalPoint(toBox2D(worldPoint)));
                },
                getLocalVector: function (worldVector) {
                    // doesn't exist in cocoon so we implement it ourselves

                    // sanity check
                    if (Utils.isDev() && !Utils.isCocoonJs()) {
                        var a = toBento(box2DBody.GetLocalVector(toBox2D(worldVector)));
                        var b = worldVector.clone().rotateRadian(-box2DBody.GetAngle());
                        if (Math.abs(a.x - b.x) > 0.00000001 || Math.abs(a.y - b.y) > 0.00000001) {
                            console.log('getLocalVector wrong result! expected:', a, 'actual:', b);
                        }
                    }

                    return worldVector.clone().rotateRadian(-box2DBody.GetAngle());
                },
                getLinearVelocityFromWorldPoint: function (worldPoint) {
                    return toBento(box2DBody.GetLinearVelocityFromWorldPoint(toBox2D(worldPoint)));
                },
                getLinearVelocityFromLocalPoint: function (localPoint) {
                    return toBento(box2DBody.GetLinearVelocityFromLocalPoint(toBox2D(localPoint)));
                },
                getLinearDamping: function () {
                    return box2DBody.GetLinearDamping();
                },
                setLinearDamping: function (linearDamping) {
                    box2DBody.SetLinearDamping(linearDamping);
                },
                getAngularDamping: function () {
                    return box2DBody.GetAngularDamping();
                },
                setAngularDamping: function (angularDamping) {
                    box2DBody.SetAngularDamping(angularDamping);
                },
                getType: function () {
                    var type = box2DBody.GetType();
                    for (var k in bodyTypes) {
                        if (bodyTypes[k] === type) {
                            return k;
                        }
                    }
                    return null;
                },
                setType: function (type) {
                    box2DBody.SetType(bodyTypes[type]);
                },
                isBullet: function () {
                    return box2DBody.IsBullet();
                },
                setBullet: function (flag) {
                    box2DBody.SetBullet(flag);
                },
                setSleepingAllowed: function (flag) {
                    box2DBody.SetSleepingAllowed(flag);
                },
                isSleepingAllowed: function () {
                    return box2DBody.IsSleepingAllowed();
                },
                setAwake: function (flag) {
                    box2DBody.SetAwake(flag);
                },
                isAwake: function () {
                    return box2DBody.IsAwake();
                },
                setFixedRotation: function (flag) {
                    box2DBody.SetFixedRotation(flag);
                },
                isFixedRotation: function () {
                    return box2DBody.IsFixedRotation();
                },
                setActive: function (flag) {
                    box2DBody.SetActive(flag);
                },
                isActive: function () {
                    return box2DBody.IsActive();
                },
                getFixtureList: function () {
                    return box2DBody.GetFixtureList();
                },
                getJointList: function () {
                    return box2DBody.GetJointList();
                },
                getControllerList: function () {
                    return box2DBody.GetControllerList();
                },
                getContactList: function () {
                    return box2DBody.GetContactList();
                },
                getNext: function () {
                    return box2DBody.GetNext();
                },

                // Userdata always refers to the Body wrapper component itself.
                // getUserData: function () {
                //     return box2DBody.GetUserData();
                // },
                // setUserData: function (data) {
                //     box2DBody.SetUserData(data);
                // },
                getWorld: function () {
                    return box2DWorld;
                }
            };

            return component;
        },

        logAvailableApi: function () {
            var out = '\nBox2D Module:\n';
            for (var k in Box2D) {
                out += k + ':\n';
                for (var k2 in Box2D[k]) {
                    out += '- ' + k2 + '\n';
                    for (var k3 in Box2D[k][k2]) {
                        out += '  |- ' + k3 + '\n';
                    }
                }
            }
            console.log(out);

            out = '\nb2Body.prototype\n';
            for (k in Box2D.Dynamics.b2Body.prototype) {
                out += '- ' + k + '\n';
            }
            console.log(out);

            out = '\nb2Fixture.prototype\n';
            for (k in Box2D.Dynamics.b2Fixture.prototype) {
                out += '- ' + k + '\n';
            }
            console.log(out);

            out = '\nb2PolygonShape\n';
            for (k in Box2D.Collision.Shapes.b2PolygonShape) {
                out += '- ' + k + '\n';
            }
            console.log(out);
        }
    };

    return Box2DModule;
});