/**
 * Module for using IronSource Ads in Bento.
 * Note: Offerwalls not implemented
 * @moduleName IronSource
 */
bento.define('modules/ironsource', [
    'bento',
    'bento/utils',
    'bento/eventsystem'
], function (
    Bento,
    Utils,
    EventSystem
) {
    var version = 2; // module version
    // Banners
    var autoLoadBanners = false;
    var bannerPrepared = false;
    var bannerLoaded = false;
    var bannerQueue = false;
    var bannerVisible = false;
    var defaultBannerOptions = {
        autoShowBanner: 'false', // true or false - To auto load ad when ad is ready.
        bannerPosition: 'bottom', // center or top or bottom. Center always overlap the screen.
        adSize: 'BANNER', // SMART OR BANNER OR LARGE_BANNER or RECT_BANNER
        placement: 'DefaultBanner', // placement name optional
        bannerOverlap: 'true', // true or false, overlap banner on webview. doesn't have effect on center position.
        offsetTopBar: 'false', // true or false, set offset when ad is loaded at top for ios7 statusbar
        offsetBottom: 'true', // true or false, set offset when
    };
    var getBannerStatus = function () {
        return {
            prepared: bannerPrepared,
            loaded: bannerLoaded,
            queue: bannerQueue,
            visible: bannerVisible
        };
    };
    var isBannerVisible = function () {
        return bannerVisible;
    };

    // Interstitials
    var autoLoadInterstitials = false;
    var isLoadingInterstitial = false;
    var isProcessingInterstitial = false;
    var interstitialAvailable = true;
    var interstitialLoadCallback;
    var interstitialLoadFailedCallback;
    var interstitialCompleteCallback;
    var interstitialFailCallback;
    var autoLoadInterstitial = function () {
        IronSource.prepareInterstitial(nothing, function () {
            // try again in 3 seconds
            window.setTimeout(autoLoadInterstitial, 3000);
        });
    };

    // Rewarded
    var rewardedAvailable = true;
    var isProcessingRewarded = false;
    var doGiveReward = false;
    var rewardedReceiveCallback;
    var rewardedCompleteCallback;
    var rewardedFailCallback;

    var nothing = function () {};
    var error = function (err) {
        if (!err) {
            return;
        }
        console.log('Iron Source Error: ', err);
    };
    var mute = function () {
        Bento.audio.muteMusic(true);
        Bento.audio.muteSound(true);
    };
    var unmute = function () {
        // restore mute state
        Bento.audio.muteMusic(Bento.saveState.load('muteMusic', false), true);
        Bento.audio.muteSound(Bento.saveState.load('muteSound', false));
        bento.require(['audia'], function (Audia) {
            var audioContext = Audia.audioContext;
            if (audioContext && audioContext.state === 'suspended') {
                audioContext.resume();
            }
        });
    };
    var Plugin;
    var IronSource = {
        // fallbacks
        version: version,
        init: nothing,
        initialize: nothing,
        setUserConsent: nothing,
        // banners
        createBannerView: nothing,
        getBannerStatus: getBannerStatus,
        isBannerVisible: isBannerVisible,
        removeBannerView: nothing,
        showBanner: nothing,
        hideBanner: nothing,
        // interstitials
        prepareInterstitial: nothing,
        canShowInterstitial: function (callback) {
            if (callback) {
                callback(false);
            }
            return false;
        },
        showInterstitial: nothing,
        // rewarded
        canShowRewarded: function (callback) {
            if (callback) {
                callback(false);
            }
            return false;
        },
        showRewarded: nothing,
        // segments
        hasSegments: function () {
            return false;
        },
        setSegment: nothing
    };

    // Check if we can access iron source, if we can, get it ready
    if (window.plugins && window.plugins.ironsource) {
        Plugin = window.plugins.ironsource;
    } else {
        //Iron Source doesn't exist, so we don't overwrite the module
        Utils.log("WARNING: IronSource not found!");
        return IronSource;
    }

    /**
     * Initializes the SDK and creates the events
     * @snippet IronSource.init()|snippet
IronSource.init({
    appKey: Utils.isNativeIos() ? '${1:<IOS_APPKEY>}' : '${2:<ANDROID_APPKEY>}',
    userConsent: true,
    userId: '', // leave empty to use default value (the IDFA)
    clientCallback: false,
    validateIntegration: true,
    autoLoadInterstitials: true,
    autoLoadBanners: true,
    onComplete: function (error) {
        if (error) {
            console.log(error);
            return;
        }
    }
});
         */
    IronSource.init = function (options) {
        var appKey = options.appKey || '';
        var userId = options.userId || '';
        var userConsent = options.userConsent;
        var validateIntegration = options.validateIntegration;
        var bannerOptions = options.bannerOptions || defaultBannerOptions;
        var onComplete = options.onComplete;
        var initPlugin = function () {
            Plugin.initSdk({
                appKey: appKey,
                userId: userId,
                clientCallback: Utils.getDefault(options.clientCallback, false),
            }, function () {
                if (autoLoadInterstitials) {
                    autoLoadInterstitial();
                }
                if (autoLoadBanners) {
                    IronSource.createBannerView(bannerOptions);
                }
                if (validateIntegration) {
                    Plugin.validateIntegration();
                }
                if (onComplete) {
                    onComplete();
                }
                console.log('Iron Source Initialized!');

            }, function (e) {
                if (onComplete) {
                    onComplete(e);
                }
            });
        };
        autoLoadInterstitials = options.autoLoadInterstitials;
        autoLoadBanners = options.autoLoadBanners;
        if (!options.appKey) {
            Utils.log('Iron Source Initialization failed!', 'No App Key!');
            return;
        }
        if (userConsent) {
            // recommended to set consent before initialization!
            IronSource.setUserConsent(true, function () {
                initPlugin();
            });
        } else {
            initPlugin();
        }


    };
    // alias for backwards compatibility
    IronSource.initialize = IronSource.init;

    /**
     * This function is used to manually set the user consent for GDPR.
     * @snippet IronSource.setUserConsent()|snippet
        IronSource.setUserConsent(true);
     */
    IronSource.setUserConsent = function (userConsent, onComplete) {
        Plugin.setUserConsent(userConsent, function () {
            console.log("IronSource: User Consent has been set to " + userConsent);
            if (onComplete) {
                onComplete();
            }
        }, function (err) {
            if (onComplete) {
                onComplete(err);
            }
            error(err);
        });
    };
    /**
     * Sets up the banner view (note: can also be set using during init if autoLoadBanners is true)
     * @snippet IronSource.createBannerView()|snippet
IronSource.createBannerView({
    autoShowBanner: 'false', // true or false - To auto load ad when ad is ready.
    bannerPosition: 'bottom', // center or top or bottom. Center always overlap the screen.
    adSize: 'BANNER', // SMART OR BANNER OR LARGE_BANNER or RECT_BANNER
    placement: 'DefaultBanner', // placement name optional
    bannerOverlap: 'true', // true or false, overlap banner on webview. doesn't have effect on center position.
    offsetTopBar: 'false', // true or false, set offset when ad is loaded at top for ios7 statusbar
    offsetBottom: 'true', // true or false, set offset when
    onComplete: function () {
        // banner creation success
    },
    onError: function (error) {
        console.log('WARNING: failed to create banner view ' + error);
    }
});
     */
    IronSource.createBannerView = function (settings) {
        var success = settings.onComplete;
        var failure = settings.onError;
        var bannerOptions = {
            autoShowBanner: settings.autoShowBanner || defaultBannerOptions.autoShowBanner, // true or false - To auto load ad when ad is ready.
            bannerPosition: settings.bannerPosition || defaultBannerOptions.bannerPosition, // center or top or bottom. Center always overlap the screen.
            adSize: settings.adSize || defaultBannerOptions.adSize, // SMART OR BANNER OR LARGE_BANNER or RECT_BANNER
            placement: settings.placement || defaultBannerOptions.placement, // placement name optional
            bannerOverlap: settings.bannerOverlap || defaultBannerOptions.bannerOverlap, // true or false, overlap banner on webview. doesn't have effect on center position.
            offsetTopBar: settings.offsetTopBar || defaultBannerOptions.offsetTopBar, // true or false, set offset when ad is loaded at top for ios7 statusbar
            offsetBottom: settings.offsetBottom || defaultBannerOptions.offsetBottom, // true or false, set offset when
        };
        Plugin.createBannerView(bannerOptions, function () {
            bannerPrepared = true;
            if (success) {
                success();
            }
        }, failure || error);
    };
    /**
     * Get status of banner
     * @snippet IronSource.getBannerStatus()|Object
        IronSource.getBannerStatus();
     */
    IronSource.getBannerStatus = getBannerStatus;
    /**
     * If banner is/was visible
     * @snippet IronSource.isBannerVisible()|Boolean
        IronSource.isBannerVisible();
     */
    IronSource.isBannerVisible = isBannerVisible;
    /**
     * Removes the banner view
     * @snippet IronSource.removeBannerView()|snippet
IronSource.removeBannerView(function () {
    // banner removeed successfully
});
     */
    IronSource.removeBannerView = function (succes) {
        Plugin.removeBannerView(function () {
            bannerPrepared = false;
            bannerLoaded = false;
            bannerQueue = false;
            bannerVisible = false;
            if (succes) {
                succes();
            }
        }, error);
    };
    /**
     * Loads and displays an ad, have to create a banner view before doing this
     * @snippet IronSource.showBanner()|snippet
IronSource.showBanner(function () {
    //success
});
     */
    IronSource.showBanner = function (succes) {
        if (bannerLoaded) {
            Plugin.showBanner(function () {
                bannerVisible = true;
                if (succes) {
                    succes();
                }
                bannerQueue = null;
            }, error);
        } else {
            // always run this since there is a difference between ad networks it seems
            IronSource.createBannerView(Utils.extend(defaultBannerOptions, {
                onComplete: function () {
                    console.log('IronSource banner view created');
                    bannerPrepared = true;
                }
            }));
            bannerQueue = succes;
        }
    };
    /**
     * Hides the current ad, doesnt remove it completely. Try removeBannerView() instead.
     * @snippet IronSource.hideBanner()|snippet
IronSource.hideBanner(function () {
    //success
});
     */
    IronSource.hideBanner = function (succes) {
        if (bannerVisible) {
            Plugin.hideBanner(function () {
                bannerVisible = false;
                if (succes) {
                    succes();
                }
            }, error);
        }
    };

    /**
     * Manually prepares an Interstital to be shown, MUST be called before showInterstitial if autoLoadInterstitials is false
     * @snippet IronSource.prepareInterstitial()|snippet
IronSource.prepareInterstitial(function () {
    //success
}, function (e) {
    //failure
});
     */
    IronSource.prepareInterstitial = function (success, failure) {
        if (interstitialAvailable) {
            // no need to load a new interstitial, return immediately
            if (success) {
                success();
            }
            return;
        }
        interstitialLoadCallback = success;
        interstitialLoadFailedCallback = failure;
        Plugin.prepareInterstitialView(nothing, error);
    };

    /**
     * Checks whether we can show an interstitial
     * @snippet IronSource.canShowInterstitial()|Boolean
IronSource.canShowInterstitial()
     */
    IronSource.canShowInterstitial = function (callback) {
        if (callback) {
            // callback for backwards compatibility
            callback(interstitialAvailable);
        }
        return interstitialAvailable;
    };
    // undocumented: async check
    IronSource.hasInterstitial = function (canShowCallback, cannotShowCallback) {
        // annoyingly this is asynchoronous
        Plugin.isInterstitialReady(function (isInterstitialReady) {
            if (isInterstitialReady) {
                canShowCallback();
            } else {
                cannotShowCallback();
            }
        }, function (e) {
            if (cannotShowCallback) {
                cannotShowCallback(e);
            } else {
                error(e);
            }
        });
    };

    /**
     * Shows an interstitial
     * @snippet IronSource.showInterstitial()|snippet
IronSource.showInterstitial({
    tag: '$1',
    muteAudio: true,
    onComplete: function (error) {
        // Interstitial video completed (or failed)
        if (error) {
            console.log('WARNING: interstitial failed to show ' + error);
        }
    }
});
     */
    IronSource.showInterstitial = function (settings) {
        var placementTag = settings.tag;
        var onComplete = settings.onComplete || error;

        if (!placementTag) {
            placementTag = '';
        }
        IronSource.canShowInterstitial(function (didSucceed) {
            if (!isProcessingInterstitial) {
                isProcessingInterstitial = true;
                interstitialCompleteCallback = function () {
                    if (settings.muteAudio) {
                        unmute();
                    }
                    onComplete();
                    // reset callbacks & bools
                    interstitialCompleteCallback = null;
                    interstitialFailCallback = null;
                    isProcessingInterstitial = false;
                    // send extra event
                    EventSystem.fire('onInterstitialAdComplete', {
                        tag: placementTag
                    });
                };
                interstitialFailCallback = function (err) {
                    if (settings.muteAudio) {
                        unmute();
                    }
                    onComplete(err || 'Failed to show');
                    // reset callbacks & bools
                    interstitialCompleteCallback = null;
                    interstitialFailCallback = null;
                    isProcessingInterstitial = false;
                };
                // mute audio before playing
                if (settings.muteAudio) {
                    mute();
                }
                Plugin.showInterstitialView(placementTag, nothing, function (err) {
                    if (settings.muteAudio) {
                        unmute();
                    }
                    console.log('Cordova error ' + err);
                    onComplete(err || 'Cordova error');
                    // reset callbacks & bools
                    interstitialCompleteCallback = null;
                    interstitialFailCallback = null;
                    isProcessingInterstitial = false;
                });
            } else {
                onComplete('Already showing interstitial');
            }
        }, function (err) {
            onComplete(err || 'Interstitial not ready to show');
        });
    };

    /**
     * Checks whether we can show an rewarded
     * @snippet IronSource.canShowRewarded()|Boolean
IronSource.canShowRewarded()
     */
    IronSource.canShowRewarded = function (callback) {
        if (callback) {
            // callback for backwards compatibility
            callback(rewardedAvailable);
        }
        return rewardedAvailable;
    };
    // undocumented: async check
    IronSource.hasRewardedVideo = function (success, failure) {
        Plugin.hasRewardedVideo(success, failure);
    };

    /**
     * Shows a rewarded ad
     * @snippet IronSource.showRewarded()|snippet
IronSource.showRewarded({
    tag: '$1',
    muteAudio: true,
    onReward: function () {
        $2
    },
    onComplete: function (error) {
        // Video completed (reward state unknown)
        $3
    }
});
     */
    IronSource.showRewarded = function (settings) {
        var onReward = settings.onReward || function () {};
        var onComplete = settings.onComplete || function () {};
        var placementTag = settings.tag || '';

        if (IronSource.canShowRewarded()) {
            if (!isProcessingRewarded) {
                isProcessingRewarded = true;
                rewardedFailCallback = function (err) {
                    if (settings.muteAudio) {
                        unmute();
                    }
                    onComplete(err || 'Failed to show rewarded ad');
                    // clear callbacks
                    rewardedCompleteCallback = null;
                    rewardedFailCallback = null;
                    isProcessingRewarded = false;
                    // also clear reward callback
                    rewardedReceiveCallback = null;
                };
                rewardedCompleteCallback = function () {
                    if (settings.muteAudio) {
                        unmute();
                    }
                    onComplete();
                    // clear callbacks
                    rewardedCompleteCallback = null;
                    rewardedFailCallback = null;
                    isProcessingRewarded = false;
                    // possibly still waiting for reward callback
                };
                rewardedReceiveCallback = function () {
                    onReward();
                    rewardedReceiveCallback = null;

                    EventSystem.fire('onRewardedAdComplete', {
                        tag: placementTag
                    });
                };
                doGiveReward = false;

                // mute audio before playing video
                if (settings.muteAudio) {
                    mute();
                }
                Plugin.showRewardedVideo(placementTag, nothing, function (e) {
                    // video failed from Cordova
                    if (settings.muteAudio) {
                        unmute();
                    }
                    onComplete(e || 'Cordova failure');

                    // clear callbacks
                    rewardedCompleteCallback = null;
                    rewardedFailCallback = null;
                    isProcessingRewarded = false;
                    // also clear reward callback
                    rewardedReceiveCallback = null;
                });
            } else {
                // we're already processing an ad
                onComplete('Already showing ad');
            }
        } else {
            // has no rewarded video
            onComplete('Rewarded ad not ready');
        }
    };
    /**
     * Has the functionality of settings segments?
     */
    IronSource.hasSegments = function () {
        return Plugin.setUserSegment;
    };
    IronSource.setSegment = function (string, callback) {
        if (!Plugin.setUserSegment) {
            if (callback) {
                callback();
            }
            return;
        }
        Plugin.setUserSegment(string, callback, callback);
    };

    /**
     * Banner events
     */
    document.addEventListener('bannerDidLoad', function (data) {
        EventSystem.fire('IronSource-bannerDidLoad', data);
        // show banner if need to
        bannerLoaded = true;
        if (bannerQueue) {
            IronSource.showBanner(bannerQueue);
        } else {}
    }, false);
    document.addEventListener('bannerDidFailToLoadWithError', function (data) {
        EventSystem.fire('IronSource-bannerDidFailToLoadWithError', data);
    }, false);
    document.addEventListener('didClickBanner', function (data) {
        EventSystem.fire('IronSource-didClickBanner', data);

    }, false);
    document.addEventListener('bannerWillPresentScreen', function (data) {
        EventSystem.fire('IronSource-bannerDidDismissScreen', data);
    }, false);
    document.addEventListener('bannerDidDismissScreen', function (data) {
        EventSystem.fire('IronSource-bannerDidDismissScreen', data);
    }, false);
    document.addEventListener('bannerWillLeaveApplication', function (data) {
        EventSystem.fire('IronSource-bannerWillLeaveApplication', data);

    }, false);

    /**
     * Interstitial events
     */
    document.addEventListener('interstitialDidLoad', function (data) {
        EventSystem.fire('IronSource-interstitialDidLoad', data);
        interstitialAvailable = true;
        if (interstitialLoadCallback) {
            interstitialLoadCallback(data);
        }
        isLoadingInterstitial = false;
    }, false);
    document.addEventListener('interstitialDidFailToLoadWithError', function (data) {
        EventSystem.fire('IronSource-interstitialDidFailToLoadWithError', data);
        interstitialAvailable = false;
        isLoadingInterstitial = false;
        if (interstitialLoadFailedCallback) {
            interstitialLoadFailedCallback(data);
        }
    }, false);

    document.addEventListener('interstitialDidShow', function (data) {
        EventSystem.fire('IronSource-interstitialDidShow', data);
    }, false);
    document.addEventListener('interstitialDidFailToShowWithError', function (data) {
        EventSystem.fire('IronSource-interstitialDidFailToShowWithError', data);
        // fire success
        if (interstitialFailCallback) {
            interstitialFailCallback(data);
        }
        // prepare next interstitial
        interstitialAvailable = false;
        if (autoLoadInterstitials) {
            autoLoadInterstitial();
        }
    }, false);
    document.addEventListener('didClickInterstitial', function (data) {
        EventSystem.fire('IronSource-didClickInterstitial', data);
    }, false);
    document.addEventListener('interstitialDidClose', function (data) {
        EventSystem.fire('IronSource-interstitialDidClose', data);
        // fire success
        if (interstitialCompleteCallback) {
            interstitialCompleteCallback(data);
        }
        // prepare next interstitial
        interstitialAvailable = false;
        if (autoLoadInterstitials) {
            autoLoadInterstitial();
        }

    }, false);
    document.addEventListener('interstitialDidOpen', function (data) {
        EventSystem.fire('IronSource-interstitialDidOpen', data);
    }, false);
    /**
     * Rewarded events
     */
    document.addEventListener('didClickRewardedVideo', function (data) {
        EventSystem.fire('IronSource-didClickRewardedVideo', data);
    }, false);
    document.addEventListener('rewardedVideoDidClose', function (data) {
        EventSystem.fire('IronSource-rewardedVideoDidClose', data);
        if (rewardedCompleteCallback) {
            rewardedCompleteCallback(data);
        }
    }, false);
    document.addEventListener('rewardedVideoDidOpen', function (data) {
        EventSystem.fire('IronSource-rewardedVideoDidOpen', data);
    }, false);
    document.addEventListener('rewardedVideoDidFailToShowWithError', function (data) {
        EventSystem.fire('IronSource-rewardedVideoDidFailToShowWithError', data);
        if (rewardedFailCallback) {
            rewardedFailCallback(data);
        }
    }, false);
    document.addEventListener('didReceiveRewardForPlacement', function (data) {
        EventSystem.fire('IronSource-didReceiveRewardForPlacement', data);
        if (rewardedReceiveCallback) {
            rewardedReceiveCallback(data);
            rewardedReceiveCallback = null;
        }
    }, false);
    document.addEventListener('rewardedVideoHasChangedAvailability', function (data) {
        EventSystem.fire('IronSource-rewardedVideoHasChangedAvailability', data);
        try {
            var decodedData = JSON.parse(data.msg);
            rewardedAvailable = decodedData.available;
        } catch (e) {
            rewardedAvailable = false;
        }
    }, false);
    
    // Note on Ironsource: "the events DidStart & DidEnd below are not available for all supported rewarded video ad networks"
    document.addEventListener('rewardedVideoDidEnd', function (data) {
        EventSystem.fire('IronSource-rewardedVideoDidEnd', data);
    }, false);
    document.addEventListener('rewardedVideoDidStart', function (data) {
        EventSystem.fire('IronSource-rewardedVideoDidStart', data);
    }, false);

    return IronSource;
});