/**
 * This module keeps track of localization json files, should be initialized after assets are loaded
 * Json files should be kept in /localization/* folder, each language json should be named by a 2 letter codename, for example en.json
 * Languages are detected by these 2 letter codes, meaning there is no distinction between regional languages, for example
 en-US and en-UK are both detected as "en" and will use the same language file.
 * One exception is Chinese. Any tradional Chinese language will be detected as "zt"
 */
bento.define('modules/localization', [
    'bento',
    'bento/utils'
], function (
    Bento,
    Utils
) {
    'use strict';
    var languageCode = 'en';
    var languages = []; // available langcodes
    var languageJson = {};
    // this function merges every json file with the same languageCode in the /localization folder
    var createLanguageJson = function (code, debug) {
        var assets = Bento.assets.getAssets();
        var jsons = assets.json;
        var conflicts = [];

        // reset the language Json
        languageJson = {};

        Utils.forEach(jsons, function (json, key, l, breakLoop) {
            if (key.indexOf('localization/') === 0 && key.slice(-2) === code) {
                if (json.base) {
                    Utils.log("WARNING: json key 'base' is a reserved name, in file " + key);
                }
                // merge the languageJson
                Utils.extend(languageJson, json, false, function (duplicateKey) {
                    var duplicate = json[duplicateKey];
                    if (Utils.isString(duplicate) || Utils.isArray(duplicate)) {
                        // strings and arrays which can't be merged
                        Utils.log('WARNING: duplicate key "' + duplicateKey + '" detected in language json "' + key + '"');
                        if (debug) {
                            debug.errors = debug.errors;
                            debug.errors += 1;
                        }
                    } else {
                        // resolve later
                        conflicts.push({
                            key: duplicateKey,
                            duplicate: duplicate,
                            original: languageJson[duplicateKey]
                        });
                    }
                });
            }
        });

        // resolve conflicts by merging the two objects, will not resolve inner conflicts
        Utils.forEach(conflicts, function (conflict, i, l, breakLoop) {
            // merge in the original
            Utils.extend(languageJson[conflict.key], conflict.original, false, function (duplicateKey) {
                Utils.log('WARNING: Inner conflict in language "' + code + '" with key "' + conflict.key + '/' + duplicateKey + '"');
                if (debug) {
                    debug.errors = debug.errors;
                    debug.errors += 1;
                }
            });
            if (debug) {
                debug.merges = debug.merges;
                debug.merges += 1;
            }
        });

        var zh_languageJson = {"loading":"加载中...","askLike":"你喜欢 {TITLE}?","comingsoon":"马上就来",
            "restore":"恢复购买","restoring":"恢复中...","restored":"您的购买已恢复！",
            "noAd":"抱歉，目前没有广告可用。","error":"哦，不，出了什么问题！",
            "errorstore":"存储错误，稍后再尝试。","purchasing":"购买中...",
            "purchased":"感谢您购买此产品。","thanks":"谢谢!","touchToStart":"触摸开始",
            "exit":"你确定退出吗?","facebook":"就像我们在Facebook上一样",
            "twitter":"在Twitter上关注我们","instagram":"关注我们的Instagram","weibo":"",
            "saveToCloud":"备份将文件保存到云?","loadFromCloud":"从云中加载保存文件?",
            "cloudSaveError":"保存到云时出现错误。请稍后再尝试。","cloudNoSaveFile":"找不到保存文件。",
            "cloudSaved":"备份将文件保存到云！","cloudLoaded":"从云中加载保存文件！",
            "cloudNoLogin":"无法登录{SERVICE}.","highscoreString":"高分",
            "muteSound":"静音","muteMusic":"静音音乐","haptics":"游戏触觉",
            "cloud":"云保存","free":"免费","removeAds":"No ads","stickers":"iMessage贴纸",
            "credits":"信用","score":"分数: {NUMBER}","best":"最好的: {NUMBER}","highscore":"新的高分！",
            "high_score":"高分","share":"分享","moreGames":"更多游戏",
            "newVersion":"有新版本可用。你想下载 %1?",
            "iapRemovesAd":"任何购买都会在游戏中删除强制广告","play":"游戏",
            "notifications?":"通知?","notificationsAsk":"你想收到通知吗?",
            "notificationMonday":"这是星期一疯狂！全天双倍收益！",
            "yesNotifications":"是","yesRatings":"是","yesCloud":"是","yesExit":"是","noRatings":"否",
            "enterName":"选择一个名字","title":"Doodle Dunk","LvX":"level {NUMBER}","skipLv":"跳过",
            "shop":"商店","unlockRandom":"随机解锁","dunked":"扣篮!","dunkText0":"令人惊叹!",
            "dunkText1":"大爆炸!","fail":"失败","next":"下一局","nextBaller":"下一个球员:","loseIt":"放弃",
            "coinBonus":"金币奖金","freeGift":"免费礼包!","new":"新!","newBall":"新球!",
            "openChests":"打开宝箱!","youGot":"你有:","topPrize":"最佳奖项:","getKeys":"获得3个钥匙",
            "unlock":"解锁","get":"获得!","tapToResume":"再次点击恢复","tapToStop":"点击停球",
            "holdToStop":"保持停球","notifications-title":"时间来扣球",
            "notifications-body0":"这些球需要扣球!",
            "notifications-body1":"只有1%可以清除这些难题！你迎接挑战吗?",
            "notifications-body2":"你好，热门镜头！团队正在等待你！",
            "btn_claim_x_amount":"领取 x {MULTIPLE}",
            "btn_claim":"领取","dailyBonus":"每日奖金！","dayNumber":"第 {NUMBER} 天"};

        return zh_languageJson;
    };
    var Localization = {
        systemLanguage: 'en',
        /**
         * Get string of current language
         * @param {String} key - key value of json
         */
        /**
         * @snippet Localization.getText()|String
        Localization.getText('$1')
         */
        /**
         * @snippet Localization.getText()|String+Replacment
        Localization.getText('$1').replace('{$2}', '$3')
         */
        getText: function (key) {
            var string = languageJson[key];
            if (!Utils.isDefined(string)) {
                Utils.log('WARNING: ' + key + ' was not found in localization strings');
            }
            return string || '';
        },
        /**
         * Change language
         * @param {String} code - language code
         */
        setLanguage: function (code) {
            code = code || Localization.systemLanguage;
            console.log('Language set to ' + code);
            try {
                languageJson = createLanguageJson(code); //Bento.assets.getJson('localization/' + code);
                languageCode = code;
            } catch (e) {
                // in case something terrible happens
                console.error('ERROR: language json ' + code + ' does not exist!!');
                console.log(e);
            }
        },
        /**
         * Inits localization with language from system/OS settings
         */
        init: function () {
            var language = window.navigator.userLanguage || window.navigator.language || 'en-US';
            var langLowerCase = language.toLowerCase();
            var code;
            var sub;
            // detect all available json files in text/
            var detectLanguages = function () {
                var assetGroups = Bento.assets.getAssetGroups();

                Utils.forEach(assetGroups, function (assetGroup) {
                    var jsons = assetGroup.json || {};
                    Utils.forEach(jsons, function (json, key, l, breakLoop) {
                        if (key.indexOf('localization/') === 0 && languages.indexOf(key.slice(-2)) < 0) {
                            // the json file must be named after a 2 letter keycode in lowercase!
                            languages.push(key.slice(-2));
                        }
                    });
                });
            };
            // compare code with detected languages
            var hasLanguage = function (langCode) {
                var i, l;
                for (i = 0, l = languages.length; i < l; ++i) {
                    if (languages[i] === langCode) {
                        return true;
                    }
                }
                return false;
            };

            console.log('Detected system language:', language);
            detectLanguages();

            // get the first 2 letters
            code = language.substr(0, 2).toLowerCase();

            // special cases for chinese tradition (zh-TW) and HK (zh-HK):
            if (
                langLowerCase === 'zh-tw' ||                // taiwan
                langLowerCase.indexOf('zh-hant') >= 0 ||    // traditional
                langLowerCase === 'zh-hk'                   // hongkong
            ) {
                code = 'zt';
            }

            // check if language json exists
            if (hasLanguage(code)) {
                languageCode = code;
            }

            // set language
            Localization.systemLanguage = languageCode;

            // for debugging purposes, create a report for every language
            if (Utils.isDev()) {
                window.inspectLanguages = function () {
                    var currentLanguageJson = languageJson;
                    Utils.forEach(languages, function (langCode, i, l, breakLoop) {
                        var debug = {
                            merges: 0,
                            errors: 0
                        };
                        var json = createLanguageJson(langCode, debug);
                        // log data
                        console.log(langCode + ': ' + Utils.getKeyLength(json) + ' keys, ' + debug.merges + ' merges,' + debug.errors + ' errors.');
                    });

                    languageJson = currentLanguageJson;
                };
            }
        },
        /**
         * Returns the current languageCode
         */
        getLanguage: function () {
            return languageCode;
        },
        /**
         * Returns true if the language was set to simplified/traditional chinese
         */
        isChinese: function () {
            return languageCode === 'zh' || languageCode === 'zt';
        },
        /**
         * Removes all languages from assets that are not the system language
         */
        cleanUnusedAssets: function () {
            var assetGroups = Bento.assets.getAssetGroups();
            Utils.forEach(assetGroups, function (assetGroup) {
                var jsons = assetGroup.json || {};
                Utils.forEach(jsons, function (json, key, l, breakLoop) {
                    var langKey = key.slice(-2);
                    if (key.indexOf('localization/') === 0 && langKey !== Localization.systemLanguage) {
                        delete jsons[key];
                    }
                });
            });
        }
    };

    return Localization;
});