/**
 * Safe wrapper for TapticEngine.js
 * https://github.com/EddyVerbruggen/cordova-plugin-taptic-engine/blob/master/www/TapticEngine.js
 * Allows haptic feedback on iPhone 7 and up
 * @moduleName TapticEngine
 */
bento.define('modules/tapticengine', [
    'bento/utils'
], function (
    Utils
) {
    'use strict';
    var nothing = function () {};
    var initUnofficial = function () {
        // this is only for iPhone 6, which doesn't have an official API, but it does contain haptic feedback
        console.log('TapticEngine: Swap official with unoficial API');
        window.TapticEngine.selection = window.TapticEngine.unofficial.weakBoom;
        window.TapticEngine.notification = window.TapticEngine.unofficial.burst;
        window.TapticEngine.impact = function (data) {
            if (!data) {
                return;
            }
            if (data.style === 'heavy') {
                window.TapticEngine.unofficial.strongBoom();
            }
            if (data.style === 'medium') {
                window.TapticEngine.unofficial.strongBoom();
            }
            if (data.style === 'light') {
                window.TapticEngine.unofficial.weakBoom();
            }
        };
    };
    var isEnabled = true;
    var module = {
        enable: function (value) {
            isEnabled = value;
        },
        unofficial: {
            weakBoom: nothing,
            strongBoom: nothing,
            burst: nothing
        },
        /**
         * Use notification feedback generators to indicate successes, failures, and warnings.
         * @snippet TapticEngine.notification()|success
TapticEngine.notification({
    type: "success"
});
         * @snippet TapticEngine.notification()|warning
TapticEngine.notification({
    type: "warning"
});
         * @snippet TapticEngine.notification()|error
TapticEngine.notification({
    type: "error"
});
         */
        notification: nothing,
        /**
         * Use selection feedback generators to indicate a change in selection.
         * @snippet TapticEngine.selection()|snippet
TapticEngine.selection();
         */
        selection: nothing,
        /**
         * Use impact feedback generators to indicate that an impact has occurred. For example, 
         * you might trigger impact feedback when a user interface object collides with something or snaps into place.
         * @snippet TapticEngine.impact()|light
TapticEngine.impact({
    type: "light"
});
         * @snippet TapticEngine.impact()|medium
TapticEngine.impact({
    type: "medium"
});
         * @snippet TapticEngine.impact()|heavy
TapticEngine.impact({
    type: "heavy"
});
         */
        impact: nothing,
        /**
         * The functions above are great for one-time events, not so much for gestures. 
         * Say for instance you want to tie this plugin to a range slider, then you 
         * can 'start' the selection first, invoke 'changed' upon changes in the range 
         * (there may be many during one gesture), then 'end' when the slider changes are done.
         * @snippet TapticEngine.gestureSelectionStart()|snippet
TapticEngine.gestureSelectionStart();
         * @snippet TapticEngine.gestureSelectionChanged()|snippet
TapticEngine.gestureSelectionChanged();
         * @snippet TapticEngine.gestureSelectionEnd()|snippet
TapticEngine.gestureSelectionEnd();
         */
        gestureSelectionStart: nothing,
        gestureSelectionChanged: nothing,
        gestureSelectionEnd: nothing,
        initUnofficial: nothing
    };
    var makeFn = function (fn) {
        return function () {
            if (isEnabled) {
                fn.apply(window.TapticEngine, arguments);
            }
        };
    };

    if (!window.TapticEngine || Utils.isNativeAndroid()) {
        console.log('WARNING: cordova plugin for TapticEngine not found.');
        return module;
    }
    console.log('Init taptic engine');
    module.initUnofficial = initUnofficial;

    // copy functions to module
    module.notification = makeFn(window.TapticEngine.notification);
    module.selection = makeFn(window.TapticEngine.selection);
    module.impact = makeFn(window.TapticEngine.impact);
    module.gestureSelectionStart = makeFn(window.TapticEngine.gestureSelectionStart);
    module.gestureSelectionChanged = makeFn(window.TapticEngine.gestureSelectionChanged);
    module.gestureSelectionEnd = makeFn(window.TapticEngine.gestureSelectionEnd);

    /**
     * Enable/disable
     * @snippet TapticEngine.active|bool
TapticEngine.active
     */
    Object.defineProperty(module, 'active', {
        get: function () {
            return isEnabled;
        },
        set: module.enable
    });

    return module;
});