/**
 * Reference-counted component. 
 *
 * @moduleName RefCount
 * @snippet RefCount|constructor
RefCount({
    id: '$1',
    constructor: function () {
        // the reference counted asset to make
        return new THREE.MeshBasicMaterial({});
    }
})
 */
bento.define('onigiri/refcount', [
    'bento',
    'bento/utils'
], function (
    Bento,
    Utils
) {
    'use strict';
    var disableRefs = false; // debugging

    // Collection of assets
    var assets = {};

    // Wrapper asset that counts down when being disposed
    var CountedAsset = function (id, asset, destructor) {
        this.id = id;
        this.count = 0;
        this.destructor = destructor || null;
        // the actual asset
        this.asset = asset;
    };
    CountedAsset.prototype.dispose = function () {
        this.count -= 1;
        if (this.count === 0) {
            if (this.destructor) {
                this.destructor(this.asset);
            } else {
                // assume the disposable asset has a "dispose function"
                this.asset.dispose();
            }
            // remove from collection
            delete assets[this.id];
        }
    };
    // shell around disposable asset, can be used as component
    var RefCount = function (settings) {
        this.name = 'refCount';
        this.id = settings.id;
        this.countedAsset = null;
        this.constructor = settings.constructor;
        this.destructor = settings.destructor;
    };
    // put the asset in use (count goes up)
    // when using this manually, make sure you actually use the asset!
    RefCount.prototype.useAsset = function () {
        var id = this.id;

        this.countedAsset = assets[id];

        // check if id exists
        if (this.countedAsset && !disableRefs) {
            // use it
            this.countedAsset.count += 1;
        } else {
            // create it for the first time
            this.countedAsset = new CountedAsset(id, this.constructor(), this.destructor);
            // default count 1?
            this.countedAsset.count = 1;
            assets[id] = this.countedAsset;
        }
        return this.countedAsset;
    };
    RefCount.prototype.dispose = function (data) {
        if (this.countedAsset) {
            this.countedAsset.dispose();
            this.countedAsset = null;
        }
    };
    // RefCount can be used as component
    RefCount.prototype.start = function (data) {
        if (!this.countedAsset) {
            this.useAsset();
        }
    };
    RefCount.prototype.destroy = function (data) {
        this.dispose();
    };

    RefCount.prototype.get = function (data) {
        if (!this.countedAsset) {
            this.useAsset();
        }
        return this.countedAsset.asset;
    };

    RefCount.assets = assets;
    RefCount.getTotalCount = function () {
        var count = 0;
        Utils.forEach(assets, function (asset) {
            count += asset.count;
        });
        return count;
    };

    RefCount.CountedAsset = CountedAsset;


    if (Bento.isDev()) {
        window.countAssets = function () {
            Utils.forEach(assets, function (asset, i, l, breakLoop) {
                console.log(i + ': ' + asset.count);
            });
        };
        window.clearRefCache = function () {
            Utils.forEach(assets, function (item, i, l, breakLoop) {
                if (item.asset && item.asset.dispose) {
                    item.asset.dispose();
                }
            });
            assets = {};
            RefCount.assets = assets;
            disableRefs = true;
        };
    }
    return RefCount;
});