/**
 * Sprite that gets rendered to the UI layer but contains geometry
 * @moduleName Sprite3d
*/
bento.define('onigiri/sprite3d', [
    'bento/utils',
    'onigiri/onigiri',
    'bento/components/sprite',
    'bento/math/vector2'
], function (
    Utils,
    Onigiri,
    Sprite,
    Vector2
) {
    'use strict';
    /* @snippet Sprite3d()|Constructor
    Sprite3d({
        object3D: ${1:null},
        width: ${2:null},
        height: ${3:null},

        // we flatten the mesh by default to not cause any issues with the rest of the UI
        //flatten: false,

        // clean up memory (geometry, materials, etc.);
        // dispose: function () {},
    });
    */
    var Sprite3d = function (settings) {
        var flatten = Utils.isDefined(settings.flatten) ? settings.flatten : true;
        this.name = settings.name || 'sprite3d';
        /**
         * Scene object, recommend against using this since the renderer applies transformations to it
         * @instance
         * @name scene
         * @type {THREE.Object3D}
         */
        this.scene = new THREE.Object3D();
        /**
         * Container object, use this to append any ThreeJS objects
         * @instance
         * @name object3D
         * @type {THREE.Object3D}
         */
        this.object3D = new THREE.Object3D();
        this.scene.add(this.object3D);
        this.object3D.scale.z = flatten ? 0.001 : 1;
        this.object3D.position.z = 0.1;
        this.flatten = settings.flatten;
        if (settings.object3D) {
            this.add(settings.object3D);
        }

        // wondering if we should set these based off the bounding box
        this.frameWidth = settings.width || 0;
        this.frameHeight = settings.height || 0;

        this.origin = new Vector2(0, 0);

        this.dispose = settings.dispose || function () {};

        Object.defineProperty(this, 'flatten', {
            get: function () {
                return flatten;
            },
            set: function (value) {
                if(flatten !== value) {
                    if (value) {
                        this.object3D.scale.z = 0.001;
                    } else {
                        this.object3D.scale.z = 1;
                    }
                    flatten = value;
                }
            }
        });

        var self = this;
        var material = {};
        var opacity = 1;
        Object.defineProperty(material, 'opacity', {
            get: function () {
                return opacity;
            },
            set: function (value) {
                opacity = value;
                self.object3D.traverse(function (obj3d) {
                    if (obj3d.material) {
                        obj3d.material.opacity = value;
                    }
                });
            }
        });
        this.material = material;
    };

    Sprite3d.prototype.start = function (data) {
        // add the parent object to the main scene
        data.renderer.three.scene.add(this.scene);
    };

    Sprite3d.prototype.destroy = function (data) {
        // remove the parent object from the main scene
        data.renderer.three.scene.remove(this.scene);
        this.dispose();
    };

    Sprite3d.prototype.draw = function (data) {
        data.renderer.render(this.scene, this.material);
    };

    Sprite3d.prototype.attached = function (data) {
        // inherit name
        this.scene.name = this.parent.name + '.' + this.name;
    };

    /**
     * Add one or an array of object3Ds to this sprite
     * @instance
     * @function
     * @name add
     * @snippet #Sprite3d.add()|add one or Array of Object3D
        add(${1})
     */
    Sprite3d.prototype.add = function (object3D) {
        // add to object3D
        if (Utils.isArray(object3D)) {
            var self = this;
            object3D.forEach(function (obj3D) {
                self.object3D.add(obj3D);
            });
        } else {
            this.object3D.add(object3D);
        }
    };

    // /**
    //  * (Internal) Clean up memory
    //  */
    // Sprite3d.prototype.dispose = function () {
    //     if (this.geometry) {
    //         this.geometry.dispose();
    //         this.geometry = null;
    //     }
    //     if (this.material) {
    //         this.material.dispose();
    //         this.material = null;
    //     }

    //     // note: textures are not disposed, they are owned by the image objects and my be reused by other instances
    // };

    Sprite3d.prototype.toString = function () {
        return '[object Sprite3d]';
    };

    Sprite3d.addToOnigiri = function () {
        Onigiri.Sprite3d = Sprite3d;
        console.log("Onigiri: added Onigiri.Sprite3d");
    };
    return Sprite3d;
});