/**
 * Extends utils with more useful functions.
 * See bento/utils
 */
bento.define('utils', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    Utils,
    Tween
) {
    'use strict';
    var utils = {
        isDebug: Utils.isDev(),
        /**
         * Wrapper for tweens to simulate a timeout
         */
        timeout: function (time, callback) {
            var tween = new Tween({
                from: 0,
                to: 1,
                in: time,
                ease: 'linear',
                onComplete: callback
            });
            return tween;
        },
        /**
         * @snippet Utils.getHashParameter()|String
        Utils.getHashParameter('${1:key}', '${2:defaultValue}')
         */
        getHashParameter: function (key, fallback, url) {
            var match;
            var result;
            url = url || window.location.hash;

            match = RegExp('[#&]' + key + '=([^&]*)').exec(url);
            result = match && decodeURIComponent(match[1].replace(/\+/g, ' '));

            if (result !== null) {
                return result;
            } else if (Utils.isDefined(fallback)) {
                return fallback;
            } else {
                return null;
            }
        },

        /**
         * Check if resolution is iPhoneX
         * @snippet Utils.isIphoneX()|Boolean
        Utils.isIphoneX()
         */
        isIphoneX: function () {
            var isIos = Utils.isNativeIos() || Utils.isApple();
            var viewport = Bento.getViewport();
            return isIos && viewport.height / viewport.width > 2;
        },
        /**
         * Get link to specific store
         */
        getStoreLink: function (device) {
            var links = {
                web: '',
                ios: '',
                android: ''
            };
            var link = links.web;
            if (!device) {
                // try to determine automatically
                if (Utils.isNativeIos()) {
                    link = links['ios'];
                } else if (Utils.isNativeAndroid()) {
                    link = links['android'];
                }
            } else if (links[device]) {
                link = links[device];
            }
            return link;
        },
        /**
         * Open Url
         */
        openUrl: function (url) {
            console.log('Opening url: ' + url);
            if (window.cordova && window.cordova.InAppBrowser) {
                window.cordova.InAppBrowser.open(url, '_system');
            } else {
                window.open(url, '_blank');
            }
        },
        /**
         * Weighted random of an array of objects. The objects must have a property called weight.
         * The randomizer will return one of the objects in the array.
         * @snippet Utils.weightedRandom()|Object
        Utils.weightedRandom(${1:arrayOfObjects})
         */
        weightedRandom: function (array) {
            // calculate sum of weights
            var weightTotal = 0;
            Utils.forEach(array, function (item, i, l, breakLoop) {
                weightTotal += item.weight || 0;
            });
            // exception: 0 total weight
            if (weightTotal === 0) {
                // assign equal weights
                Utils.forEach(array, function (item) {
                    item.weight = 1;
                });
                weightTotal = array.length;
            }

            // randomize a number
            var random = Utils.getRandomRangeFloat(0, weightTotal);
            // find the winner
            var winner;
            var weightIncr = 0;
            Utils.forEach(array, function (item, i, l, breakLoop) {
                if (item.weight <= 0) {
                    return;
                }
                weightIncr += item.weight;
                if (weightIncr > random) {
                    winner = item;
                    breakLoop();
                    return;
                }
            });
            return winner;
        },
    };
    Utils.extend(Utils, utils);
    return Utils;
});